import re
from collections import OrderedDict
from enum import Enum
from functools import partialmethod, wraps
from pathlib import Path
from typing import Dict, Optional, Union

from ._utils import _chained, _get_enum_str


# --- metadata -----------------------------------------------------------------
class MetadataMixin:
    """Derived class can have metadata assigned to it as key value pairs."""

    @_chained
    def add_metadata(self, *args: Dict[str, Union[str, int, float, bool]], **kwargs):
        """
        add_metadata(self, *args: Dict[str, Union[str, int, float, bool]], **kwargs)
        Add metadata key value pairs to this object

        .. code-block:: python

            # Example 1
            job.add_metadata({"key1": "value1"})

            # Example 2
            job.add_metadata(key1="value1, key2="value2")

        :param args: dictionary of key value pair to add as metadata
        :type args: Dict[str, Union[str, int, float, bool]]
        :raises TypeError: each arg in args must be a dict
        :return: self
        """

        # values will be converted to str; metadata schema requires str values
        for arg in args:
            if isinstance(arg, dict):
                self.metadata.update({key: str(value) for key, value in arg.items()})
            else:
                raise TypeError(f"{arg} must be of type dict")

        for key, value in kwargs.items():
            self.metadata[key] = str(value)


# --- hooks --------------------------------------------------------------------
class EventType(Enum):
    """Event type on which a hook will be triggered"""

    NEVER = "never"
    START = "start"
    ERROR = "error"
    SUCCESS = "success"
    END = "end"
    ALL = "all"


class HookMixin:
    """Derived class can have hooks assigned to it. This currently supports
    shell hooks. The supported hooks are triggered when some event,
    specified by :py:class:`~Pegasus.api.mixins.EventType`, takes place.
    """

    @_chained
    def add_shell_hook(self, event_type: EventType, cmd: str):
        """
        add_shell_hook(self, event_type: EventType, cmd: str)
        Add a shell hook. The given command will be executed by the shell
        when the specified :py:class:`~Pegasus.api.mixins.EventType` takes
        place.

        .. code-block:: python

            # Example
            wf.add_shell_hook(EventType.START, "echo 'hello'")

        :param event_type: an event type defined in :py:class:`~Pegasus.api.mixins.EventType`
        :type event_type: EventType
        :param cmd: shell command
        :type cmd: str
        :raises ValueError: event_type must be one of :py:class:`~Pegasus.api.mixins.EventType`
        :return: self
        """
        # TODO: consider making event_type either an event type or an actual ShellHook
        if not isinstance(event_type, EventType):
            raise ValueError("event_type must be one of EventType")

        self.hooks[_ShellHook.__hook_type__].append(_ShellHook(event_type, cmd))


class _Hook:
    """Base class that specific hook types will inherit from"""

    def __init__(self, event_type):
        """Constructor

        :param event_type: an event type defined in :py:class:`~Pegasus.api.mixins.EventType`
        :type event_type: EventType
        :raises ValueError: event_type must be of type :py:class:`~Pegasus.api.mixins.EventType`
        """
        if not isinstance(event_type, EventType):
            raise ValueError("event_type must be one of EventType")

        self.on = event_type.value

    # TODO: def get/set event type


# TODO: make this public
class _ShellHook(_Hook):
    """A hook that executes a shell command"""

    __hook_type__ = "shell"

    def __init__(self, event_type, cmd):
        """Constructor

        :param event_type: an event type defined in :py:class:`~Pegasus.api.mixins.EventType`
        :type event_type: EventType
        :param cmd: shell command
        :type cmd: str
        """
        _Hook.__init__(self, event_type)
        self.cmd = cmd

    def __json__(self):
        return OrderedDict([("_on", self.on), ("cmd", self.cmd)])


# --- profiles -----------------------------------------------------------------
class Namespace(Enum):
    """Profile Namespace values recognized by Pegasus"""

    PEGASUS = "pegasus"
    CONDOR = "condor"
    DAGMAN = "dagman"
    ENV = "env"
    GLOBUS = "globus"
    SELECTOR = "selector"
    STAT = "stat"


def _profiles(ns, **map_p):
    """Internal decorator that enables the use of kw args in functions like
    ProfileMixin.add_condor_profile() and ProfileMixin.add_dagman_profile().
    A handful of profile keys contain "." or "-", and so those profile keys cannot
    be used for kw args. By providing a mapping of legal key names to actual key names,
    adding profiles becomes more natural. For example we can have the following:

    .. code-block:: python

        # Example
        @_profiles(
            Namespace.DAGMAN,
            pre_args="PRE.ARGUMENTS"
        )
        def add_dagman_profile(self, pre_args: str = None):
            ...

    This way, available Profile keys will appear in an IDE and we can use kw args
    for keys that would be invalid as python variable names.

    :param ns: namespace
    :type ns: Namespace
    """

    def wrap(f):
        @wraps(f)
        def wrapped_f(self, *a, **kw):
            # map of the actual property name to a value
            # for example, "abort_dag" would become {"ABORT-DAG": some_value}
            new_kw = {}
            for k, v in kw.items():
                if v is not None and k in map_p:
                    if isinstance(map_p[k], str):
                        new_kw[map_p[k]] = v
                    elif isinstance(map_p[k], (tuple, list)):
                        # for some properties, we can pass a long a function as well
                        # that can be used to do thing such as convert from one unit
                        nk, vf = map_p[k]
                        new_kw[nk] = vf(v)

            # calling this will verify kwargs
            f(self, **kw)

            return ProfileMixin.add_profiles(self, ns, **new_kw)

        return wrapped_f

    return wrap


def to_mb(value: str) -> int:
    """Convert the given value to MB

    :param value: str formatted as str formatted as :code:`'<int> [MB | GB | TB | PB | EB | ZB | YB]'`
    :type value: str
    :raises ValueError: invalid format
    :return: value in MB
    :rtype: int
    """
    try:
        value = str(value).strip()
        m = re.match(r"^\s*(\d+([Ee][+-]?\d+)?)\s*([MmGgPpEeZzYy][Bb])?\s*$", value)

        if not m:
            raise ValueError

        # float(..) is required as int("1E1") raises a ValueError
        amt = int(float(m.group(1)))
        unit = m.group(3).lower() if m.group(3) else None

        if unit is None:
            return amt
        else:
            _bytes = {
                "mb": 1 << 20,
                "gb": 1 << 30,
                "tb": 1 << 40,
                "pb": 1 << 50,
                "eb": 1 << 60,
                "zb": 1 << 70,
                "yb": 1 << 80,
            }

            return int((amt * _bytes[unit]) / _bytes["mb"])
    except Exception:
        raise ValueError(
            "value: {} should be a str formatted as '<int> [MB | GB | TB | PB | EB | ZB | YB]'".format(
                value
            )
        )


def to_kb(value: str) -> int:
    """Convert the given value to KB

    :param value: str formatted as str formatted as :code:`'<int> [MB | GB | TB | PB | EB | ZB | YB]'`
    :type value: str
    :raises ValueError: invalid format
    :return: value in KB
    :rtype: int
    """
    return to_mb(value) * 1024


class ProfileMixin:
    @_chained
    def add_profiles(
        self,
        ns: Namespace,
        key: Optional[str] = None,
        value: Optional[Union[str, int, float, bool, Path]] = None,
        **kw,
    ):
        r"""
        add_profiles(self, ns: Namespace, key: Optional[str] = None, value: Optional[str, int, float, bool, Path] = None, **kw)
        Add a profile.

        If key and value are given, then :code:`**kw` are ignored and :code:`{Namespace::key : value}`
        is added. Else :code:`**kw` is added. When the value of "key" is not a valid python
        variable name, the usage in Example #1 should be used, else follow the usage
        shown in Example #2.

        .. code-block:: python

            # Example 1
            job.add_profiles(Namespace.DAGMAN, key="pre.arguments", value="-i f1")

            # Example 2
            job.add_profiles(Namespace.ENV, JAVA_HOME="/usr/bin/java", USER="ryan")

        For :py:meth:`~Pegasus.api.mixins.ProfileMixin.add_globus_profiles`, :py:meth:`~Pegasus.api.mixins.ProfileMixin.add_condor_profiles`,
        :py:meth:`~Pegasus.api.mixins.ProfileMixin.add_dagman_profiles`, :py:meth:`~Pegasus.api.mixins.ProfileMixin.add_selector_profiles`,
        and :py:meth:`~Pegasus.api.mixins.ProfileMixin.add_pegasus_profiles`, if a profile key
        you are trying to use is not listed as a key word argument, use this
        function to add the profile.

        :raises TypeError: namespace must be one of Namespace
        :return: self
        """

        if not isinstance(ns, Namespace):
            raise TypeError(
                "invalid ns: {ns}; ns should be one of {enum_str}".format(
                    ns=ns, enum_str=_get_enum_str(Namespace)
                )
            )

        ns = ns.value

        # add profile(s)
        if key and value:
            if isinstance(value, Path) or type(value) == bool:
                # convert pathlib.Path and bool to str
                value = str(value)

            self.profiles[ns].update({key: value})
        else:
            for k, v in kw.items():
                if isinstance(v, Path) or type(v) == bool:
                    # convert pathlib.Path, bool, to str
                    kw[k] = str(v)

            self.profiles[ns].update(kw)

    #: Add environment variable(s)
    add_env = partialmethod(add_profiles, Namespace.ENV)

    @_profiles(
        Namespace.GLOBUS,
        count="count",
        job_type="jobtype",
        max_cpu_time="maxcputime",
        max_memory=("maxmemory", to_mb),
        max_time="maxtime",
        max_wall_time="maxwalltime",
        min_memory=("minmemory", to_mb),
        project="project",
        queue="queue",
    )
    def add_globus_profile(
        self,
        *,
        count: int = None,
        job_type: str = None,
        max_cpu_time: int = None,
        max_memory: str = None,
        max_time: int = None,
        max_wall_time: int = None,
        min_memory: int = None,
        project: str = None,
        queue: str = None,
    ):
        """Add Globus profile(s).

        The globus profile namespace encapsulates Globus resource specification
        language (RSL) instructions. The RSL configures settings and behavior of
        the remote scheduling system.

        :param count: the number of times an executable is started, defaults to None
        :type count: int, optional
        :param job_type: specifies how the job manager should start the remote job. While Pegasus defaults to single, use mpi when running MPI jobs., defaults to None
        :type job_type: str, optional
        :param max_cpu_time: the max CPU time in minutes for a single execution of a job, defaults to None
        :type max_cpu_time: int, optional
        :param max_memory: the maximum memory in MB required for the job. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type max_memory: str, optional
        :param max_time: the maximum time or walltime in minutes for a single execution of a job, defaults to None
        :type max_time: int, optional
        :param max_wall_time: the maximum walltime in minutes for a single execution of a job, defaults to None
        :type max_wall_time: int, optional
        :param min_memory: the minumum amount of memory required for this job. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type min_memory: str, optional
        :param project: associates an account with a job at the remote end, defaults to None
        :type project: str, optional
        :param queue: the remote queue in which the job should be run. Used when remote scheduler is PBS that supports queues, defaults to None
        :type queue: str, optional
        :return: self
        """
        ...

    @_profiles(
        Namespace.CONDOR,
        universe="universe",
        periodic_release="periodic_release",
        periodic_remove="periodic_remove",
        filesystem_domain="filesystemdomain",
        stream_error="stream_error",
        stream_output="stream_output",
        priority="priority",
        request_cpus="request_cpus",
        request_gpus="request_gpus",
        request_memory=("request_memory", to_mb),
        request_disk=("request_disk", to_kb),
        requirements="requirements",
        should_transfer_files="should_transfer_files",
        when_to_transfer_output="when_to_transfer_output",
        condor_collector="condor_collector",
        grid_resource="grid_resource",
        cream_attributes="cream_attributes",
    )
    def add_condor_profile(
        self,
        *,
        universe: str = None,
        periodic_release: str = None,
        periodic_remove: str = None,
        filesystem_domain: str = None,
        stream_error: bool = None,
        stream_output: bool = None,
        priority: str = None,
        request_cpus: str = None,
        request_gpus: str = None,
        request_memory: str = None,
        request_disk: str = None,
        requirements: str = None,
        should_transfer_files: str = None,
        when_to_transfer_output: str = None,
        condor_collector: str = None,
        grid_resource: str = None,
        cream_attributes: str = None,
    ):
        """Add Condor profile(s).

        The condor profiles permit to add or overwrite instructions in the Condor submit file.

        :param universe: Pegasus defaults to either globus or scheduler universes. Set to standard for compute jobs that require standard universe. Set to vanilla to run natively in a condor pool, or to run on resources grabbed via condor glidein, defaults to None
        :type universe: str, optional
        :param periodic_release: is the number of times job is released back to the queue if it goes to HOLD, e.g. due to Globus errors. Pegasus defaults to 3, defaults to None
        :type periodic_release: str, optional
        :param periodic_remove: is the number of times a job is allowed to get into HOLD state before being removed from the queue. Pegasus defaults to 3, defaults to None
        :type periodic_remove: str, optional
        :param filesystem_domain: Useful for Condor glide-ins to pin a job to a remote site, defaults to None
        :type filesystem_domain: str, optional
        :param stream_error: boolean to turn on the streaming of the stderr of the remote job back to submit host, defaults to None
        :type stream_error: bool, optional
        :param stream_output: boolean to turn on the streaming of the stdout of the remote job back to submit host, defaults to None
        :type stream_output: bool, optional
        :param priority: integer value to assign the priority of a job. Higher value means higher priority. The priorities are only applied for vanilla / standard/ local universe jobs. Determines the order in which a users own jobs are executed, defaults to None
        :type priority: str, optional
        :param request_cpus: Number of CPU's a job requires, defaults to None
        :type request_cpus: str, optional
        :param request_gpus: Number of GPU's a job requires, defaults to None
        :type request_gpus: str, optional
        :param request_memory: Amount of memory a job requires. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type request_memory: str, optional
        :param request_disk: Amount of disk a job requires. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type request_disk: str, optional
        :param requirements: a job requirements expression such as :code:`"(CUDACapability >= 1.2) && $(requirements:True)"`
        :type requirements: str, optional
        :param should_transfer_files: Used to define if HTCondor should transfer files to and from the remote machine where the job runs, Given as :code:`"YES"`, :code:`"NO"`, or :code:`"IF_NEEDED"`
        :type should_transfer_file: str, optional
        :param when_to_transfer_output: Given as a one of :code:`"ON_EXIT"`, :code:`"ON_EXIT_OR_EVICT"`, or :code:`"ON_SUCCESS"`
        :type when_to_transfer_output: str, optional
        :param condor_collector: Specify the condor collector to use (e.g :code:`"ccg-testing999.isi.edu"`)
        :type condor_collector: str, optional
        :param grid_resource: Specify a grid resource such as :code:`"batch pbs"`
        :type grid_resource: str, optional,
        :param cream_attributes: Additional cream attributes (e.g. :code:`"key1=value1;key2=value2"`)
        :type cream_attributes: str, optional
        :return: self
        """
        ...

    @_profiles(
        Namespace.PEGASUS,
        clusters_num="clusters.num",
        clusters_size="clusters.size",
        job_aggregator="job.aggregator",
        job_aggregator_arguments="job.aggregator.arguments",
        grid_start="gridstart",
        grid_start_path="gridstart.path",
        grid_start_arguments="gridstart.arguments",
        grid_start_launcher="gridstart.launcher",
        grid_start_launcher_arguments="gridstart.launcher.arguments",
        stagein_clusters="stagein.clusters",
        stagein_local_clusters="stagein.local.clusters",
        stagein_remote_clusters="stagein.remote.clusters",
        stageout_clusters="stageout.clusters",
        stageout_local_clusters="stageout.local.clusters",
        stageout_remote_clusters="stageout.remote.clusters",
        group="group",
        change_dir="change.dir",
        create_dir="create.dir",
        transfer_proxy="transfer.proxy",
        style="style",
        pmc_request_memory=("pmc_request_memory", to_mb),
        pmc_request_cpus="pmc_request_cpus",
        pmc_priority="pmc_priority",
        pmc_task_arguments="pmc_task_arguments",
        exitcode_failure_msg="exitcode.failuremsg",
        exitcode_success_msg="exitcode.successmsg",
        checkpoint_time="checkpoint.time",
        max_walltime="maxwalltime",
        glite_arguments="glite.arguments",
        auxillary_local="auxillary.local",
        condor_arguments_quote="condor.arguments.quote",
        runtime="runtime",
        clusters_max_runtime="clusters.maxruntime",
        cores="cores",
        gpus="gpus",
        nodes="nodes",
        ppn="ppn",
        memory=("memory", to_mb),
        diskspace=("diskspace", to_mb),
        data_configuration="data.configuration",
        queue="queue",
        project="project",
        boto_config="BOTO_CONFIG",
        container_arguments="container.arguments",
        container_launcher="container.launcher",
        container_launcher_arguments="container.launcher.arguments",
        label="label",
        pegasus_lite_env_source="pegasus_lite_env_source",
        SSH_PRIVATE_KEY="SSH_PRIVATE_KEY",
        relative_submit_dir="relative.submit.dir",
    )
    def add_pegasus_profile(
        self,
        *,
        clusters_num: int = None,
        clusters_size: int = None,
        job_aggregator: int = None,
        job_aggregator_arguments: str = None,
        grid_start: int = None,
        grid_start_path: str = None,
        grid_start_arguments: str = None,
        grid_start_launcher: str = None,
        grid_start_launcher_arguments: str = None,
        stagein_clusters: int = None,
        stagein_local_clusters: int = None,
        stagein_remote_clusters: int = None,
        stageout_clusters: int = None,
        stageout_local_clusters: int = None,
        stageout_remote_clusters: int = None,
        group: str = None,
        change_dir: bool = None,
        create_dir: bool = None,
        transfer_proxy: bool = None,
        style: str = None,
        pmc_request_memory: str = None,
        pmc_request_cpus: int = None,
        pmc_priority: int = None,
        pmc_task_arguments: str = None,
        exitcode_failure_msg: str = None,
        exitcode_success_msg: str = None,
        checkpoint_time: int = None,
        max_walltime: int = None,
        glite_arguments: str = None,
        auxillary_local: bool = None,
        condor_arguments_quote: bool = None,
        runtime: str = None,
        clusters_max_runtime: int = None,
        cores: int = None,
        gpus: int = None,
        nodes: int = None,
        ppn: int = None,
        memory: str = None,
        diskspace: str = None,
        data_configuration=None,
        queue: str = None,
        project: str = None,
        boto_config: str = None,
        container_arguments: str = None,
        container_launcher: str = None,
        container_launcher_arguments: str = None,
        label: str = None,
        pegasus_lite_env_source: Union[str, Path] = None,
        SSH_PRIVATE_KEY: str = None,
        relative_submit_dir: Union[str, Path] = None,
    ):
        """Add Pegasus profile(s).

        :param clusters_num: Determines the total number of clusters per level, jobs are evenly spread across clusters (see `Pegasus Clustering Guide <https://pegasus.isi.edu/documentation/job_clustering.php#horizontal_clustering>`_ for more information), defaults to None
        :type clusters_num: int, optional
        :param clusters_size: Determines the number of jobs in each cluster (see `Pegasus Clustering Guide <https://pegasus.isi.edu/documentation/job_clustering.php#horizontal_clustering>`_ for more information), defaults to None
        :type clusters_size: int, optional
        :param job_aggregator: Indicates the clustering executable that is used to run the clustered job on the remote site, defaults to None
        :type job_aggregator: int, optional
        :param job_aggregator_arguments: Additional arguments with which a clustering executable should be invoked, defaults to None
        :type job_aggregator_arguments: str, optional
        :param grid_start: Determines the executable for launching a job (see `docs <https://pegasus.isi.edu/documentation/profiles.php#hints_profiles>`__ for more information), defaults to None
        :type grid_start: int, optional
        :param grid_start_path: Sets the path to the gridstart . This profile is best set in the Site Catalog, defaults to None
        :type grid_start_path: str, optional
        :param grid_start_arguments: Sets the arguments with which GridStart is used to launch a job on the remote site, defaults to None
        :type grid_start_arguments: str, optional
        :param grid_start_launcher: specifies the path to the executable to launch kickstart, defaults to None,
        :type grid_start_launcher: str, optional
        :param grid_start_launcher_arguments: the arguments to the launcher executable if any
        :type grid_start_launcher_arguments: str, optional
        :param stagein_clusters: This key determines the maximum number of stage-in jobs that are can executed locally or remotely per compute site per workflow. This is used to configure the BalancedCluster Transfer Refiner, which is the Default Refiner used in Pegasus. This profile is best set in the Site Catalog or in the Properties file, defaults to None
        :type stagein_clusters: int, optional
        :param stagein_local_clusters: This key provides finer grained control in determining the number of stage-in jobs that are executed locally and are responsible for staging data to a particular remote site. This profile is best set in the Site Catalog or in the Properties file, defaults to None
        :type stagein_local_clusters: int, optional
        :param stagein_remote_clusters: This key provides finer grained control in determining the number of stage-in jobs that are executed remotely on the remote site and are responsible for staging data to it. This profile is best set in the Site Catalog or in the Properties file, defaults to None
        :type stagein_remote_clusters: int, optional
        :param stageout_clusters: This key determines the maximum number of stage-out jobs that are can executed locally or remotely per compute site per workflow. This is used to configure the BalancedCluster Transfer Refiner, , which is the Default Refiner used in Pegasus, defaults to None
        :type stageout_clusters: int, optional
        :param stageout_local_clusters: This key provides finer grained control in determining the number of stage-out jobs that are executed locally and are responsible for staging data from a particular remote site. This profile is best set in the Site Catalog or in the Properties file, defaults to None
        :type stageout_local_clusters: int, optional
        :param stageout_remote_clusters: This key provides finer grained control in determining the number of stage-out jobs that are executed remotely on the remote site and are responsible for staging data from it. This profile is best set in the Site Catalog or in the Properties file, defaults to None
        :type stageout_remote_clusters: int, optional
        :param group: Tags a job with an arbitrary group identifier. The group site selector makes use of the tag, defaults to None
        :type group: str, optional
        :param change_dir: If true, tells kickstart to change into the remote working directory. Kickstart itself is executed in whichever directory the remote scheduling system chose for the job, defaults to None
        :type change_dir: bool, optional
        :param create_dir: If true, tells kickstart to create the the remote working directory before changing into the remote working directory. Kickstart itself is executed in whichever directory the remote scheduling system chose for the job, defaults to None
        :type create_dir: bool, optional
        :param transfer_proxy: If true, tells Pegasus to explicitly transfer the proxy for transfer jobs to the remote site. This is useful, when you want to use a full proxy at the remote end, instead of the limited proxy that is transferred by CondorG, defaults to None
        :type transfer_proxy: bool, optional
        :param style: Sets the condor submit file style. If set to globus, submit file generated refers to CondorG job submissions. If set to condor, submit file generated refers to direct Condor submission to the local Condor pool. It applies for glidein, where nodes from remote grid sites are glided into the local condor pool. The default style that is applied is globus, defaults to None
        :type style: str, optional
        :param pmc_request_memory: This key is used to set the -m option for pegasus-mpi-cluster. It specifies the amount of memory in MB that a job requires. This profile is usually set in the DAX for each job. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type pmc_request_memory: str, optional
        :param pmc_request_cpus: This key is used to set the -c option for pegasus-mpi-cluster. It specifies the number of cpu's that a job requires. This profile is usually set in the DAX for each job, defaults to None
        :type pmc_request_cpus: int, optional
        :param pmc_priority: This key is used to set the -p option for pegasus-mpi-cluster. It specifies the priority for a job . This profile is usually set in the DAX for each job. Negative values are allowed for priorities, defaults to None
        :type pmc_priority: int, optional
        :param pmc_task_arguments: The key is used to pass any extra arguments to the PMC task during the planning time. They are added to the very end of the argument string constructed for the task in the PMC file. Hence, allows for overriding of any argument constructed by the planner for any particular task in the PMC job, defaults to None
        :type pmc_task_arguments: str, optional
        :param exitcode_failure_msg: The message string that pegasus-exitcode searches for in the stdout and stderr of the job to flag failures, defaults to None
        :type exitcode_failure_msg: str, optional
        :param exitcode_success_msg: The message string that pegasus-exitcode searches for in the stdout and stderr of the job to determine whether a job logged it's success message or not. Note this value is used to check for whether a job failed or not i.e if this profile is specified, and pegasus-exitcode DOES NOT find the string in the job stdout or stderr, the job is flagged as failed. The complete rules for determining failure are described in the man page for pegasus-exitcode, defaults to None
        :type exitcode_success_msg: str, optional
        :param checkpoint_time: the expected time in minutes for a job after which it should be sent a TERM signal to generate a job checkpoint file, defaults to None
        :type checkpoint_time: int, optional
        :param max_walltime: the maximum walltime in minutes for a single execution of a job, defaults to None
        :type max_walltime: int, optional
        :param glite_arguments: specifies the extra arguments that must appear in the local PBS generated script for a job, when running workflows on a local cluster with submissions through Glite. This is useful when you want to pass through special options to underlying LRMS such as PBS e.g. you can set value -l walltime=01:23:45 -l nodes=2 to specify your job's resource requirements, defaults to None
        :type glite_arguments: str, optional
        :param auxillary_local: indicates whether auxillary jobs associated with a compute site X, can be run on local site. This CAN ONLY be specified as a profile in the site catalog and should be set when the compute site filesystem is accessible locally on the submit host, defaults to None
        :type auxillary_local: bool, optional
        :param condor_arguments_quote: indicates whether condor quoting rules should be applied for writing out the arguments key in the condor submit file. By default it is true unless the job is schedule to a glite style site. The value is automatically set to false for glite style sites, as condor quoting is broken in batch_gahp, defaults to None
        :type condor_arguments_quote: bool, optional
        :param runtime: Specifies the expected runtime of a job in seconds, defaults to None
        :type runtime: str, optional
        :param clusters_max_runtime: Specifies the maximum runtime of a job, defaults to None
        :type clusters_max_runtime: int, optional
        :param cores: The total number of cores required for a job. This is also used for accounting purposes in the database while generating statistics. It corresponds to the multiplier_factor in the job_instance table, defaults to None
        :type cores: int, optional
        :param gpus: The total number of gpus required for a job, defaults to None
        :type gpus: int, optional
        :param nodes: Indicates the number of nodes a job requires, defaults to None
        :type nodes: int, optional
        :param ppn: Indicates the number of processors per node. This profile is best set in the Site Catalog and usually set when running workflows with MPI jobs, defaults to None
        :type ppn: int, optional
        :param memory: Indicates the maximum memory a job requires. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type memory: str, optional
        :param diskspace: Indicates the maximum diskspace a job requires in MB. Given as a str formatted as '<int> [MB | GB | TB | PB | EB]', defaults to None
        :type diskspace: int, optional
        :param data_configuration: Indicates the data configuration setup. Can be one of "sharedfs", "condorio", or "nonsharedfs", defaults to None
        :type data_configuration: str, optional
        :param queue: This specifies the queue for the job. (e.g. :code:`"normal"`)
        :type queue: str, optional
        :param project: Causes the job time to be charged to or associated with a particular project/account. This is not used for SGE.
        :type project: str, optional
        :param boto_config: Specified which :code:`.boto` file to use (e.g. :code:`"/home/myuser/.boto"`)
        :type boto_config: str, optional
        :param container_arguments: additional cli arguments that will be added to the :code:`docker container run` or :code:`singularity exec` command
        :type container_arguments: str, optional
        :param container_launcher: wrapper to launch the container invocation in PegasusLite for running the container e.g.  :code:`srun singularity exec` command
        :type container_launcher: str, optional
        :param container_launcher_arguments: additional cli arguments that will be added to the wrapper to launch the container invocation in PegasusLite for running the container e.g.  :code:`srun singularity exec` command
        :type container_launcher_arguments: str, optional
        :param label: associate a label to a job; used for label based clustering
        :type label: str, optional
        :param pegasus_lite_env_source: specify a path on the submit host to indicate the file that needs to be sourced, defaults to None
        :type pegasus_lite_env_source: Union[str, Path], optional
        :param SSH_PRIVATE_KEY: path to the ssh private key which will be used in this workflow (it is recommended that a special set of keys are created specifically for workflow use)
        :type SSH_PRIVATE_KEY: str, optional
        :param relative_submit_dir: specify the relative directory a job's submit files are written to, defaults to None
        :type relative_submit_dir: Union[str, Path], optional
        """
        ...

    @_profiles(
        Namespace.SELECTOR,
        execution_site="execution.site",
        pfn="pfn",
        grid_job_type="grid.jobtype",
    )
    def add_selector_profile(
        self,
        *,
        execution_site: str = None,
        pfn: Union[str, Path] = None,
        grid_job_type: str = None,
    ):
        """Add Selector(s).

        The Selector namespace allows users to override the beahvior of the Workflow
        Mapper during site selection. This gives you finer grained control over
        where a job executes and what executable it refers to.

        :param execution_site: the execution site where a job should be executed, defaults to None
        :type execution_site: str, optional
        :param pfn: the physical file name to the main executable that a job refers to. Overrides any entries specified in the transformation catalog, defaults to None
        :type pfn: Union[str, Path], optional
        :param grid_job_type: This profile is usually used to ensure that a compute job executes on another job manager (see `docs <https://pegasus.isi.edu/documentation/profiles.php#hints_profiles>`__ for more information), defaults to None
        :type grid_job_type: str, optional
        :return: self
        """
        ...

    @_profiles(
        Namespace.DAGMAN,
        pre="PRE",
        pre_arguments="PRE.ARGUMENTS",
        post="POST",
        post_arguments="POST.ARGUMENTS",
        retry="RETRY",
        category="CATEGORY",
        priority="PRIORITY",
        abort_dag_on="ABORT-DAG-ON",
        max_pre="MAXPRE",
        max_post="MAXPOST",
        max_jobs="MAXJOBS",
        max_idle="MAXIDLE",
        post_scope="POST.SCOPE",
    )
    def add_dagman_profile(
        self,
        *,
        pre: Union[str, Path] = None,
        pre_arguments: str = None,
        post: str = None,
        post_arguments: str = None,
        retry: str = None,
        category: str = None,
        priority: str = None,
        abort_dag_on: str = None,
        max_pre: str = None,
        max_post: str = None,
        max_jobs: str = None,
        max_idle: str = None,
        post_scope: str = None,
    ):
        """Add Dagman profile(s).

        Note, to add the profile keys "post.path.[value of dagman.post]" or
        ["category-name].maxjobs", use the following:

        .. code-block:: python

            job.add_profiles(Namespace.DAGMAN, key="post.path.<value of dagman.post>", value=<value string>)
            job.add_profiles(Namespace.DAGMAN, key="<category-name>.maxjobs", value=<value string>)

        :param pre: is the path to the pre-script. DAGMan executes the pre-script before it runs the job, defaults to None
        :type pre: Union[str, Path], optional
        :param pre_arguments: are command-line arguments for the pre-script, if any, defaults to None
        :type pre_arguments: str, optional
        :param post: is the postscript type/mode that a user wants to associate with a job (see `docs <https://pegasus.isi.edu/documentation/profiles.php>`__ for more information), defaults to None
        :type post: str, optional
        :param post_arguments: are the command line arguments for the post script, if any, defaults to None
        :type post_arguments: str, optional
        :param retry: is the number of times DAGMan retries the full job cycle from pre-script through post-script, if failure was detected, defaults to None
        :type retry: int, optional
        :param category: the DAGMan category the job belongs to, defaults to None
        :type category: str, optional
        :param priority: the priority to apply to a job. DAGMan uses this to select what jobs to release when MAXJOBS is enforced for the DAG, defaults to None
        :type priority: int, optional
        :param abort_dag_on: The ABORT-DAG-ON key word provides a way to abort the entire DAG if a given node returns a specific exit code (AbortExitValue). The syntax for the value of the key is AbortExitValue [RETURN DAGReturnValue] . When a DAG aborts, by default it exits with the node return value that caused the abort. This can be changed by using the optional RETURN key word along with specifying the desired DAGReturnValue, defaults to None
        :type abort_dag_on: str, optional
        :param max_pre: sets the maximum number of PRE scripts within the DAG that may be running at one time, defaults to None
        :type max_pre: str, optional
        :param max_post: sets the maximum number of POST scripts within the DAG that may be running at one time, defaults to None
        :type max_post: str, optional
        :param max_jobs: sets the maximum number of jobs within the DAG that will be submitted to Condor at one time, defaults to None
        :type max_jobs: str, optional
        :param max_idle: Sets the maximum number of idle jobs allowed before HTCondor DAGMan stops submitting more jobs. Once idle jobs start to run, HTCondor DAGMan will resume submitting jobs. If the option is omitted, the number of idle jobs is unlimited, defaults to None
        :type max_idle: str, optional
        :param post_scope: can be "all", "none" or "essential" (see `docs <https://pegasus.isi.edu/documentation/profiles.php>`__ for more information), defaults to None
        :type post_scope: str, optional
        :return: self
        """
        ...

    # initially had add_<namespace>_profile for 5.0 release and in order not not
    # break existing workflows, these functions have aliases that are plural
    add_globus_profiles = add_globus_profile
    add_pegasus_profiles = add_pegasus_profile
    add_dagman_profiles = add_dagman_profile
    add_condor_profiles = add_condor_profile
    add_selector_profiles = add_selector_profile



import getpass
import json
import os
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryFile

import pytest
import yaml

from Pegasus.api.errors import PegasusError
from Pegasus.api.writable import Writable, _CustomEncoder, _filter_out_nones


@pytest.fixture(scope="function")
def writable_obj():
    def _writable_obj():
        class Item:
            def __init__(self, n):
                self.name = "item" + str(n)

            def __json__(self):
                return {"name": self.name}

        class Container(Writable):
            _DEFAULT_FILENAME = "container.yml"

            def __init__(self):
                Writable.__init__(self)

                self.name = "container⽷"
                self.items = [Item(i) for i in range(3)]

            def __json__(self):
                return {"name": self.name, "items": self.items}

        return Container()

    return _writable_obj()


@pytest.fixture(scope="function")
def expected():
    return {
        "x-pegasus": {
            "createdBy": getpass.getuser(),
            "createdOn": "now",
            "apiLang": "python",
        },
        "name": "container⽷",
        "items": [{"name": "item0"}, {"name": "item1"}, {"name": "item2"}],
    }


class Test_CustomEncoder:
    def test_json(self, writable_obj, expected):
        result = json.loads(json.dumps(writable_obj, cls=_CustomEncoder))

        # delete file info as it is only included when object is written
        # but not converted to json
        del expected["x-pegasus"]

        assert result == expected

    def test_json_attribute_not_callable(self, writable_obj):
        writable_obj.__json__ = "not a callable"
        with pytest.raises(TypeError) as e:
            json.loads(json.dumps(writable_obj, cls=_CustomEncoder))

        assert "__json__ is not callable" in str(e)

    def test_json_default_encoder(self, mocker):
        class NotJSONSerializableObj:
            pass

        # default implementation json.JSONEncoder.default should raise TypeError
        with pytest.raises(TypeError) as e:
            json.loads(json.dumps(NotJSONSerializableObj(), cls=_CustomEncoder))

        # ensure we are not getting the TypeError raised by _CustomEncoder.default
        assert "__json__ is not callable" not in str(e)


class TestWritable:
    def test_write_using_defaults(self, writable_obj, expected, mocker):
        writable_obj.write()
        with open("container.yml") as f:
            result = yaml.safe_load(f)

            # TODO: check that keys in x-pegasus exist and then del x-pegasus
            # setting dates to be the same as it won't be safe to compare them
            # and this is simpler than trying to mocker.patch datetime
            expected["x-pegasus"]["createdOn"] = result["x-pegasus"]["createdOn"]
            assert result == expected

        os.remove("container.yml")

    @pytest.mark.parametrize(
        "file, _format, loader",
        [
            ("file.yml", "yml", yaml.safe_load),
            ("file.yaml", "yaml", yaml.safe_load),
            ("file", "yml", yaml.safe_load),
            ("file.123", "yml", yaml.safe_load),
            ("file.yml", "json", yaml.safe_load),
            ("file.json", "json", json.load),
            ("file", "json", json.load),
            ("file.123", "json", json.load),
        ],
    )
    def test_write_using_str_input(self, writable_obj, expected, file, _format, loader):
        writable_obj.write(file, _format=_format)
        with open(file) as f:
            result = loader(f)

            # setting dates to be the same as it won't be safe to compare them
            # and this is simpler than trying to mocker.patch datetime
            expected["x-pegasus"]["createdOn"] = result["x-pegasus"]["createdOn"]
            assert result == expected

        os.remove(file)

    @pytest.mark.parametrize(
        "file, _format, loader",
        [
            ("file.yml", "yml", yaml.safe_load),
            ("file.yaml", "yaml", yaml.safe_load),
            ("file", "yml", yaml.safe_load),
            ("file.123", "yml", yaml.safe_load),
            ("file.yml", "json", yaml.safe_load),
            ("file.json", "json", json.load),
            ("file", "json", json.load),
            ("file.123", "json", json.load),
        ],
    )
    def test_write_using_file_input(
        self, writable_obj, expected, file, _format, loader
    ):
        with open(file, "w+") as f:
            writable_obj.write(f, _format=_format)
            f.seek(0)
            result = loader(f)

            # setting dates to be the same as it won't be safe to compare them
            # and this is simpler than trying to mocker.patch datetime
            expected["x-pegasus"]["createdOn"] = result["x-pegasus"]["createdOn"]
            assert result == expected

        os.remove(file)

    def test_write_invalid_format(self, writable_obj):
        with pytest.raises(ValueError):
            writable_obj.write("abc", _format="123")

    def test_write_invalid_file_type(self, writable_obj):
        with pytest.raises(TypeError) as e:
            writable_obj.write(1)

        assert "1 must be of type str or file object" in str(e)

    @pytest.mark.parametrize(
        "file, _format, loader",
        [
            (TemporaryFile, "yml", yaml.safe_load),
            (TemporaryFile, "json", json.load),
            (StringIO, "yaml", yaml.safe_load),
        ],
    )
    def test_write_stream_without_name(
        self, writable_obj, expected, file, _format, loader
    ):

        kwargs = {"mode": "w+"}
        if type(file) == type(StringIO):
            kwargs = {}

        with file(**kwargs) as f:
            writable_obj.write(f, _format=_format)
            f.seek(0)
            result = loader(f)

            # setting dates to be the same as it won't be safe to compare them
            # and this is simpler than trying to mocker.patch datetime
            expected["x-pegasus"]["createdOn"] = result["x-pegasus"]["createdOn"]
            assert result == expected

    def test_internal_write_invalid_format(self, writable_obj):
        """Testing Writable._write"""
        with pytest.raises(ValueError) as e:
            writable_obj._write("file", "bad_format")

        assert "invalid _ext: bad_format" in str(e)

    def test_set_path_given_default_filename(self, writable_obj):
        writable_obj.write()
        assert writable_obj._path == str(Path(writable_obj._DEFAULT_FILENAME).resolve())
        assert writable_obj.path == Path(writable_obj._DEFAULT_FILENAME).resolve()
        writable_obj._path = None
        os.remove(str(Path(writable_obj._DEFAULT_FILENAME).resolve()))

    def test_set_path_given_str_filename(self, writable_obj):
        writable_obj.write("filename")
        assert writable_obj._path == str(Path("filename").resolve())
        assert writable_obj.path == Path("filename").resolve()
        writable_obj._path = None
        os.remove("filename")

    def test_set_path_given_NamedTemporaryFile(self, writable_obj):
        f = NamedTemporaryFile(mode="w")
        writable_obj.write(f)
        assert writable_obj._path == str(Path(f.name).resolve())
        assert writable_obj.path == Path(f.name).resolve()
        writable_obj._path = None
        f.close()

    def test_set_path_given_TemporaryFile(self, writable_obj):
        f = TemporaryFile(mode="w")
        writable_obj.write(f)
        assert writable_obj._path == None
        with pytest.raises(PegasusError) as e:
            writable_obj.path

        assert "Container.write(filename)" in str(e)
        writable_obj._path = None
        f.close()

    def test_set_path_given_StringIO(self, writable_obj):
        f = StringIO()
        writable_obj.write(f)
        assert writable_obj._path == None
        with pytest.raises(PegasusError) as e:
            writable_obj.path

        assert "Container.write(filename)" in str(e)
        f.close()

    def test_get_path_exception_message(self, writable_obj):
        writable_obj.write(StringIO())

        with pytest.raises(PegasusError) as e:
            writable_obj.path

        assert "Container.write(filename)" in str(e)


def test_filter_out_nones():
    d = {"a": 1, "b": None}

    assert _filter_out_nones(d) == {"a": 1}


def test_filter_out_nones_invalid_type():
    with pytest.raises(TypeError) as e:
        _filter_out_nones(123)



from collections import defaultdict
from pathlib import Path

import pytest

from Pegasus.api.mixins import (
    EventType,
    HookMixin,
    MetadataMixin,
    Namespace,
    ProfileMixin,
    _Hook,
    _ShellHook,
    to_mb,
)


@pytest.fixture(scope="function")
def md_mixin_obj():
    def _metadata_mixin_obj():
        class MetadataMixinObj(MetadataMixin):
            def __init__(self):
                self.metadata = dict()

        return MetadataMixinObj()

    return _metadata_mixin_obj()


class TestMetadataMixin:
    def test_add_metadata_method_1(self, md_mixin_obj):
        assert id(md_mixin_obj.add_metadata({"key": "value"})) == id(md_mixin_obj)
        assert md_mixin_obj.metadata["key"] == "value"

    def test_add_metadata_method_2(self, md_mixin_obj):
        assert id(md_mixin_obj.add_metadata(key1="value1", key2="value2")) == id(
            md_mixin_obj
        )
        assert md_mixin_obj.metadata["key1"] == "value1"
        assert md_mixin_obj.metadata["key2"] == "value2"

    def test_add_invalid_metadata(self, md_mixin_obj):
        with pytest.raises(TypeError) as e:
            md_mixin_obj.add_metadata([1, 2])

        assert "[1, 2]" in str(e)

    def test_chaining(self, md_mixin_obj):
        assert md_mixin_obj.add_metadata(key1="value").add_metadata(key2="value")

        assert len(md_mixin_obj.metadata) == 2
        assert id(md_mixin_obj) == id(md_mixin_obj.add_metadata(key3="value"))


class Test_Hook:
    def test_valid_hook(self):
        _Hook(EventType.START)

    def test_invalid_hook(self):
        with pytest.raises(ValueError):
            _Hook("123")


class Test_ShellHook:
    def test_valid_shell_hook(self):
        _ShellHook(EventType.START, "some command")

    def test_invalid_shell_hook(self):
        with pytest.raises(ValueError):
            _ShellHook("123", "some command")

    def test_tojson(self):
        hook = _ShellHook(EventType.START, "some command")
        assert hook.__json__() == {"_on": "start", "cmd": "some command"}


@pytest.fixture(scope="function")
def hook_mixin_obj():
    def _hook_mixin_obj():
        class HookMixinObj(HookMixin):
            def __init__(self):
                self.hooks = defaultdict(list)

        return HookMixinObj()

    return _hook_mixin_obj()


class TestHookMixin:
    def test_add_shell_hook(self, hook_mixin_obj):
        cmd = "some command"
        hook_mixin_obj.add_shell_hook(EventType.START, cmd)
        assert "shell" in hook_mixin_obj.hooks

        added_shell_hook = hook_mixin_obj.hooks["shell"][0]

        assert added_shell_hook.on == EventType.START.value
        assert added_shell_hook.cmd == cmd

    def test_add_invalid_shell_hook(self, hook_mixin_obj):
        with pytest.raises(ValueError):
            hook_mixin_obj.add_shell_hook("123", "some command")

    def test_chaining(self, hook_mixin_obj):
        (
            hook_mixin_obj.add_shell_hook(
                EventType.START, "some command"
            ).add_shell_hook(EventType.START, "another command")
        )

        assert "shell" in hook_mixin_obj.hooks

        added_shell_hook_1 = hook_mixin_obj.hooks["shell"][0]
        assert added_shell_hook_1.on == EventType.START.value
        assert added_shell_hook_1.cmd == "some command"

        added_shell_hook_2 = hook_mixin_obj.hooks["shell"][1]
        assert added_shell_hook_2.on == EventType.START.value
        assert added_shell_hook_2.cmd == "another command"

        assert id(hook_mixin_obj) == id(
            hook_mixin_obj.add_shell_hook(EventType.START, "some command")
        )


@pytest.fixture(scope="function")
def obj():
    def _profile_mixin_obj():
        class ProfileMixinObj(ProfileMixin):
            def __init__(self):
                self.profiles = defaultdict(dict)

        return ProfileMixinObj()

    return _profile_mixin_obj()


class TestProfileMixin:
    @pytest.mark.parametrize(
        "given_value,expected_value",
        [
            (1, 1),
            (1.0, 1.0),
            ("value", "value"),
            (True, "True"),
            (Path("/file/path.txt"), "/file/path.txt"),
        ],
    )
    def test_add_valid_profile(self, obj, given_value, expected_value):
        assert id(
            obj.add_profiles(Namespace.ENV, key="ABC-123.45/", value=given_value)
        ) == id(obj)
        assert dict(obj.profiles) == {"env": {"ABC-123.45/": expected_value}}

    @pytest.mark.parametrize(
        "given_value,expected_value",
        [
            (1, 1),
            (1.0, 1.0),
            ("value", "value"),
            (False, "False"),
            (Path("/file/path.txt"), "/file/path.txt"),
        ],
    )
    def test_add_valid_profiles(self, obj, given_value, expected_value):
        assert id(
            obj.add_profiles(Namespace.ENV, ENV1="env1", ENV2=given_value,)
        ) == id(obj)

        assert dict(obj.profiles) == {"env": {"ENV1": "env1", "ENV2": expected_value}}

    def test_add_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_profiles("ns")

        assert "invalid ns: ns" in str(e)

    def test_add_env(self, obj):
        assert id(obj.add_env(ENV1="env1", ENV2="env2")) == id(obj)
        assert dict(obj.profiles) == {"env": {"ENV1": "env1", "ENV2": "env2"}}

    def test_add_globus_profile(self, obj):
        assert id(
            obj.add_globus_profile(
                count=1,
                job_type="single",
                max_cpu_time=2,
                max_memory="3",
                max_time=4,
                max_wall_time=5,
                min_memory="2 GB",
                project="abc",
                queue="queue",
            )
        ) == id(obj)

        assert dict(obj.profiles) == {
            "globus": {
                "count": 1,
                "jobtype": "single",
                "maxcputime": 2,
                "maxmemory": 3,
                "maxtime": 4,
                "maxwalltime": 5,
                "minmemory": 2048,
                "project": "abc",
                "queue": "queue",
            }
        }

    def test_add_globus_profiles(self, obj):
        assert id(obj.add_globus_profiles(count=1)) == id(obj)
        assert dict(obj.profiles) == {"globus": {"count": 1}}

    def test_add_globus_profile_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_globus_profile(aa=1)

        assert "add_globus_profile() got an unexpected" in str(e)

    def test_add_dagman_profile(self, obj):
        assert id(
            obj.add_dagman_profile(
                pre="pre",
                pre_arguments="pre_args",
                post="post",
                post_arguments="post_args",
                retry=1,
                category="cat",
                priority="prio",
                abort_dag_on="abrt",
                max_pre="mp",
                max_post="mp",
                max_jobs="mj",
                max_idle="mi",
                post_scope="ps",
            )
        ) == id(obj)

        assert dict(obj.profiles) == {
            "dagman": {
                "PRE": "pre",
                "PRE.ARGUMENTS": "pre_args",
                "POST": "post",
                "POST.ARGUMENTS": "post_args",
                "RETRY": 1,
                "CATEGORY": "cat",
                "PRIORITY": "prio",
                "ABORT-DAG-ON": "abrt",
                "MAXPRE": "mp",
                "MAXPOST": "mp",
                "MAXJOBS": "mj",
                "MAXIDLE": "mi",
                "POST.SCOPE": "ps",
            }
        }

    def test_add_dagman_profiles(self, obj):
        assert id(obj.add_dagman_profiles(pre="pre")) == id(obj)
        assert dict(obj.profiles) == {"dagman": {"PRE": "pre"}}

    def test_add_dagman_profile_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_dagman_profile(aa=1)

        assert "add_dagman_profile() got an unexpected" in str(e)

    def test_add_condor_profile(self, obj):
        assert id(
            obj.add_condor_profile(
                universe="un",
                periodic_release="pr",
                periodic_remove="pr",
                filesystem_domain="fsd",
                stream_error="se",
                stream_output="so",
                priority="prio",
                request_cpus="rc",
                request_gpus="rg",
                request_memory="100 MB",
                request_disk="200 MB",
                requirements="(CUDACapability >= 1.2) && $(requirements:True)",
                should_transfer_files="YES",
                when_to_transfer_output="ON_EXIT",
                condor_collector="ccg-testing999.isi.edu",
                grid_resource="batch pbs",
                cream_attributes="key1=value1",
            )
        ) == id(obj)

        assert dict(obj.profiles) == {
            "condor": {
                "universe": "un",
                "periodic_release": "pr",
                "periodic_remove": "pr",
                "filesystemdomain": "fsd",
                "stream_error": "se",
                "stream_output": "so",
                "priority": "prio",
                "request_cpus": "rc",
                "request_gpus": "rg",
                "request_memory": 100,
                "request_disk": 204800,
                "requirements": "(CUDACapability >= 1.2) && $(requirements:True)",
                "should_transfer_files": "YES",
                "when_to_transfer_output": "ON_EXIT",
                "condor_collector": "ccg-testing999.isi.edu",
                "grid_resource": "batch pbs",
                "cream_attributes": "key1=value1",
            }
        }

    def test_add_condor_profiles(self, obj):
        assert id(obj.add_condor_profiles(universe="vanilla")) == id(obj)
        assert dict(obj.profiles) == {"condor": {"universe": "vanilla"}}

    def test_add_condor_profile_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_condor_profile(aa=1)

        assert "add_condor_profile() got an unexpected" in str(e)

    def test_add_pegasus_profile(self, obj):
        assert id(
            obj.add_pegasus_profile(
                clusters_num="clusters.num",
                clusters_size="clusters.size",
                job_aggregator="job.aggregator",
                job_aggregator_arguments="job.aggregator.arguments",
                grid_start="gridstart",
                grid_start_path="gridstart.path",
                grid_start_arguments="gridstart.arguments",
                stagein_clusters="stagein.clusters",
                stagein_local_clusters="stagein.local.clusters",
                stagein_remote_clusters="stagein.remote.clusters",
                stageout_clusters="stageout.clusters",
                stageout_local_clusters="stageout.local.clusters",
                stageout_remote_clusters="stageout.remote.clusters",
                group="group",
                change_dir="change.dir",
                create_dir="create.dir",
                transfer_proxy="transfer.proxy",
                style="style",
                pmc_request_memory="512",
                pmc_request_cpus="pmc_request_cpus",
                pmc_priority="pmc_priority",
                pmc_task_arguments="pmc_task_arguments",
                exitcode_failure_msg="exitcode.failuremsg",
                exitcode_success_msg="exitcode.successmsg",
                checkpoint_time=1,
                max_walltime="maxwalltime",
                glite_arguments="glite.arguments",
                auxillary_local="auxillary.local",
                condor_arguments_quote="condor.arguments.quote",
                runtime="runtime",
                clusters_max_runtime="clusters.maxruntime",
                cores=1,
                gpus=1,
                nodes="nodes",
                ppn="ppn",
                memory="2 GB",
                diskspace="1 GB",
                data_configuration="condorio",
                queue="normal",
                project="-A project_name",
                boto_config="/home/myuser/.boto",
                container_arguments="--shm-size",
                container_launcher="/usr/bin/srun",
                container_launcher_arguments="--kill-on-bad-exit",
                label="nodeLabel",
                pegasus_lite_env_source="/path",
                SSH_PRIVATE_KEY="/path",
                relative_submit_dir="some-dir",
            )
        ) == id(obj)

        assert dict(obj.profiles) == {
            "pegasus": {
                "clusters.num": "clusters.num",
                "clusters.size": "clusters.size",
                "job.aggregator": "job.aggregator",
                "job.aggregator.arguments": "job.aggregator.arguments",
                "gridstart": "gridstart",
                "gridstart.path": "gridstart.path",
                "gridstart.arguments": "gridstart.arguments",
                "stagein.clusters": "stagein.clusters",
                "stagein.local.clusters": "stagein.local.clusters",
                "stagein.remote.clusters": "stagein.remote.clusters",
                "stageout.clusters": "stageout.clusters",
                "stageout.local.clusters": "stageout.local.clusters",
                "stageout.remote.clusters": "stageout.remote.clusters",
                "group": "group",
                "change.dir": "change.dir",
                "create.dir": "create.dir",
                "transfer.proxy": "transfer.proxy",
                "style": "style",
                "pmc_request_memory": 512,
                "pmc_request_cpus": "pmc_request_cpus",
                "pmc_priority": "pmc_priority",
                "pmc_task_arguments": "pmc_task_arguments",
                "exitcode.failuremsg": "exitcode.failuremsg",
                "exitcode.successmsg": "exitcode.successmsg",
                "checkpoint.time": 1,
                "maxwalltime": "maxwalltime",
                "glite.arguments": "glite.arguments",
                "auxillary.local": "auxillary.local",
                "condor.arguments.quote": "condor.arguments.quote",
                "runtime": "runtime",
                "clusters.maxruntime": "clusters.maxruntime",
                "cores": 1,
                "gpus": 1,
                "nodes": "nodes",
                "ppn": "ppn",
                "memory": 2048,
                "diskspace": 1024,
                "data.configuration": "condorio",
                "queue": "normal",
                "project": "-A project_name",
                "BOTO_CONFIG": "/home/myuser/.boto",
                "container.arguments": "--shm-size",
                "container.launcher": "/usr/bin/srun",
                "container.launcher.arguments": "--kill-on-bad-exit",
                "label": "nodeLabel",
                "pegasus_lite_env_source": "/path",
                "SSH_PRIVATE_KEY": "/path",
                "relative.submit.dir": "some-dir",
            }
        }

    def test_add_pegasus_profiles(self, obj):
        assert id(obj.add_pegasus_profiles(gpus=1)) == id(obj)
        assert dict(obj.profiles) == {"pegasus": {"gpus": 1}}

    def test_add_pegasus_profile_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_pegasus_profile(aa=1)

        assert "add_pegasus_profile() got an unexpected" in str(e)

    def test_add_selector_profile(self, obj):
        assert id(
            obj.add_selector_profile(
                execution_site="condor-pool", pfn="/tmp", grid_job_type="compute"
            )
        ) == id(obj)

        assert dict(obj.profiles) == {
            "selector": {
                "execution.site": "condor-pool",
                "pfn": "/tmp",
                "grid.jobtype": "compute",
            }
        }

    def test_add_selector_profiles(self, obj):
        assert id(obj.add_selector_profiles(pfn="/tmp")) == id(obj)
        assert dict(obj.profiles) == {"selector": {"pfn": "/tmp"}}

    def test_add_selector_profile_invalid_profile(self, obj):
        with pytest.raises(TypeError) as e:
            obj.add_selector_profile(aa=1)

        assert "add_selector_profile() got an unexpected" in str(e)


@pytest.mark.parametrize(
    "value, expected",
    [
        ("0", 0),
        (1, 1),
        ("1", 1),
        ("2 MB", 2),
        ("2 GB", 2048),
        ("10 GB", 10240),
        ("1MB", 1),
        ("1mb", 1),
        ("1E2", 100),
        ("100E-2", 1),
        ("1e2   MB", 100),
        ("100e-2   Eb", 1099511627776),
        ("1   zB   ", 1125899906842624),
        ("  1   yb", 1.152921504606847e18),
    ],
)
def test_to_mb(value, expected):
    assert to_mb(value) == expected


@pytest.mark.parametrize(
    "value", [("abc MB"), ("MB"), ("1 KB"), "1M", "1m", "0.5E2", "1e-1000    Mm", "MB"]
)
def test_to_mb_invalid_input(value):
    with pytest.raises(ValueError) as e:
        to_mb(value)

    assert f"value: {value} should be a str" in str(e)



import json
import os
import re

import pytest
import yaml

from Pegasus.api.writable import _CustomEncoder


def _tojson(obj):
    """Returns dict representation of obj using writable._CustomEncoder"""
    return json.loads(json.dumps(obj, cls=_CustomEncoder))


@pytest.fixture(scope="module")
def convert_yaml_schemas_to_json():
    """
    Convert all the yaml schemas into json files.
    These files will be used whenever schema validation is
    needed for the tests. The json files will be cleaned up
    at the end of the test module.
    """
    # get the path of the schema file with the given name
    path = os.path.dirname(os.path.realpath(__file__))
    path = path.replace("packages/pegasus-api/test/api", "share/pegasus/schema/yaml")

    json_schemas = {
        os.path.join(path, filename): os.path.join(
            path, filename.replace(".yml", ".json")
        )
        for filename in [
            "common.yml",
            "rc-5.0.yml",
            "tc-5.0.yml",
            "sc-5.0.yml",
            "wf-5.0.yml",
        ]
    }

    # convert each of the yml schemas to json
    for yml_filename, json_filename in json_schemas.items():
        with open(yml_filename) as yml_file, open(json_filename, "w") as json_file:

            json_str = json.dumps(yaml.safe_load(yml_file))

            # for references pointing to '*.yml' files, convert them to point
            # to '.json' files instead
            json_str = re.sub(
                r"([a-z0-9\-\.]+)(.yml)",
                os.path.join("file://" + path, r"\1.json"),
                json_str,
            )

            json_obj = json.loads(json_str)
            json_obj["$id"] = "file://" + json_filename

            json.dump(json_obj, json_file, indent=4)

    yield

    # cleanup
    for _, json_schema in json_schemas.items():
        os.remove(json_schema)


@pytest.fixture(scope="function")
def load_schema():
    """
    Load the given schema into memory.
    """

    def _load_schema(name):
        # get the path of the schema file with the given name
        path = os.path.dirname(os.path.realpath(__file__))
        path = path.replace(
            "packages/pegasus-api/test/api", "share/pegasus/schema/yaml"
        )
        path = os.path.join(path, name)

        with open(path) as f:
            return json.load(f)

    return _load_schema



from .mixins import EventType, Namespace
from .properties import Properties
from .replica_catalog import File, ReplicaCatalog
from .site_catalog import (
    OS,
    Arch,
    Directory,
    FileServer,
    Grid,
    Operation,
    Scheduler,
    Site,
    SiteCatalog,
    SupportedJobs,
)
from .transformation_catalog import (
    Container,
    Transformation,
    TransformationCatalog,
    TransformationSite,
)
from .workflow import Job, SubWorkflow, Workflow

from Pegasus.client._client import PegasusClientError



import pytest
from jsonschema import Draft7Validator

# Validating the actual format of the schemas against the Draft 7 meta-schema


@pytest.mark.parametrize(
    "schema_file",
    [
        ("common.json"),
        ("wf-5.0.json"),
        ("rc-5.0.json"),
        ("tc-5.0.json"),
        ("sc-5.0.json"),
    ],
)
def test_schema(schema_file, convert_yaml_schemas_to_json, load_schema):
    schema = load_schema(schema_file)
    Draft7Validator.check_schema(schema)

    # everything defined under $defs is not automatically validated unless we
    # explicitly do so
    for _, _def in schema["$defs"].items():
        Draft7Validator.check_schema(_def)



import getpass
import json
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Optional, TextIO, Union

from .errors import PegasusError

from Pegasus import yaml

__all__ = ["Writable"]


class _CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        # TODO: handle instance of Date and Path
        """
        if isinstance(obj, Date):
                        return "whatever spec we come up with for Date such as ISO8601"
        elif isinstance(obj, Path):
            return obj.resolve
        """

        if hasattr(obj, "__json__"):
            if callable(obj.__json__):
                return obj.__json__()
            else:
                raise TypeError(f"__json__ is not callable for {obj}")

        return json.JSONEncoder.default(self, obj)


def _filter_out_nones(_dict):
    """Helper function to remove keys where their values are set to None to avoid cluttering yaml/json files

    :param _dict: object represented as a dict
    :type _dict: dict
    :raises ValueError: _dict must be of type dict
    :return: new dictionary with 'None' values removed
    :rtype: dict
    """
    if not isinstance(_dict, dict):
        raise TypeError(f"invalid _dict: {_dict}; _dict must be of type {type(dict)}")

    return OrderedDict([(k, v) for k, v in _dict.items() if v is not None])


class Writable:
    """Derived class can be serialized to a json or yaml file"""

    _FORMATS = {"yml", "yaml", "json"}

    def __init__(self):
        self._path = None

    @property
    def path(self) -> Path:
        """
        Retrieve the path to which this object has been written to.

        :raises PegasusError: object has not yet been written to a file
        :return: resolved path to which this object has been written
        :rtype: Path
        """
        if self._path == None:
            raise PegasusError(
                "{}.write(filename) must be called before trying to retrieve path".format(
                    self.__class__.__name__
                )
            )

        return Path(self._path)

    def _write(self, file, _format):
        """Internal function to dump to file in either yaml or json formats

        :param file: file object to write to
        :type file: file
        :param _format: file format that can be "yml", "yaml", or "json
        :type _ext: str
        :raises ValueError: _format must be one of "yml", "yaml" or "json"
        """
        if _format.lower() not in Writable._FORMATS:
            raise ValueError(
                "invalid _ext: {_format}, extension must be one of {formats}".format(
                    _format=_format, formats=Writable._FORMATS
                )
            )

        # add file info
        self_as_dict = OrderedDict(
            [
                (
                    "x-pegasus",
                    {
                        "createdBy": getpass.getuser(),
                        "createdOn": datetime.now().strftime(r"%m-%d-%yT%H:%M:%SZ"),
                        "apiLang": "python",
                    },
                )
            ]
        )

        self_as_dict.update(
            json.loads(
                json.dumps(self, cls=_CustomEncoder), object_pairs_hook=OrderedDict
            )
        )

        if _format == "yml" or _format == "yaml":
            # TODO: figure out how to get yaml.dump to recurse down into nested objects
            # yaml.dump(_CustomEncoder().default(self), file, sort_keys=False)
            yaml.dump(self_as_dict, file, allow_unicode=True)
        else:
            json.dump(
                self_as_dict, file, cls=_CustomEncoder, indent=4, ensure_ascii=False
            )

    def write(self, file: Optional[Union[str, TextIO]] = None, _format: str = "yml"):
        """Serialize this class as either yaml or json and write to the given
        file. If file==None, this class will be written to a default file. The
        following classes have these defaults:

        .. table:: Default Files
            :widths: auto

            =====================  ===================
            Class                  Default Filename
            =====================  ===================
            SiteCatalog            sites.yml
            ReplicaCatalog         replicas.yml
            TransformationCatalog  transformations.yml
            Workflow               workflow.yml
            =====================  ===================

        :param file: path or file object (opened in "w" mode) to write to, defaults to None
        :type file: Optional[Union[str, TextIO]]
        :param _format: can be either "yml", "yaml" or "json", defaults to "yml"
        :type _format: str, optional
        :raises ValueError: _format must be one of "yml", "yaml" or "json"
        :raises TypeError: file must be a str or file object
        """
        if _format.lower() not in Writable._FORMATS:
            raise ValueError(
                "invalid file format: {_format}, format should be one of 'yml', 'yaml', or 'json'"
            )

        # default file name
        if file is None:
            file = self._DEFAULT_FILENAME

        # do the write
        if isinstance(file, str):
            path = Path(file)
            ext = path.suffix[1:].lower()

            with open(file, "w") as f:
                if ext in Writable._FORMATS:
                    self._write(f, ext)
                else:
                    self._write(f, _format)

            self._path = str(path.resolve())

        elif hasattr(file, "read"):
            try:
                f = Path(str(file.name))
                ext = f.suffix[1:]
            except AttributeError:
                # writing to a stream such as StringIO with
                # no attr "name"
                self._write(file, _format)
            else:
                if ext in Writable._FORMATS:
                    self._write(file, ext)
                else:
                    self._write(file, _format)

                if isinstance(file.name, str):
                    self._path = str(Path(file.name).resolve())

        else:
            raise TypeError(f"{file} must be of type str or file object")



import getpass
import json
import re
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
import yaml
from jsonschema import validate

from Pegasus.api.errors import DuplicateError
from Pegasus.api.site_catalog import (
    OS,
    PEGASUS_VERSION,
    Arch,
    Directory,
    FileServer,
    Grid,
    Operation,
    Scheduler,
    Site,
    SiteCatalog,
    SupportedJobs,
)
from Pegasus.api.writable import _CustomEncoder


class TestFileServer:
    def test_valid_file_server(self):
        assert FileServer("url", Operation.PUT)

    def test_invlaid_file_server(self):
        with pytest.raises(TypeError) as e:
            FileServer("url", "put")

        assert "invalid operation_type: put" in str(e)

    def test_tojson_with_profiles(self, convert_yaml_schemas_to_json, load_schema):
        file_server = FileServer("url", Operation.PUT).add_env(SOME_ENV="1")

        result = json.loads(json.dumps(file_server, cls=_CustomEncoder))

        expected = {
            "url": "url",
            "operation": "put",
            "profiles": {"env": {"SOME_ENV": "1"}},
        }

        file_server_schema = load_schema("sc-5.0.json")["$defs"]["fileServer"]
        validate(instance=result, schema=file_server_schema)

        assert result == expected


class TestDirectory:
    def test_valid_directory(self):
        assert Directory(Directory.LOCAL_SCRATCH, "/path")
        assert Directory(
            Directory.SHARED_SCRATCH, Path("/abs/path"), shared_file_system=True
        )

    def test_invalid_directory_type(self):
        with pytest.raises(TypeError) as e:
            Directory("invalid type", "/path")

        assert "invalid directory_type: invalid type" in str(e)

    def test_invalid_directory_path(self):
        with pytest.raises(ValueError) as e:
            Directory(Directory.LOCAL_SCRATCH, Path("dir"))

        assert "invalid path: {}".format(Path("dir")) in str(e)

    def test_add_valid_file_server(self):
        d = Directory(Directory.LOCAL_SCRATCH, "/path")
        assert d.add_file_servers(FileServer("url", Operation.PUT))

    def test_add_invalid_file_server(self):
        with pytest.raises(TypeError) as e:
            d = Directory(Directory.LOCAL_SCRATCH, "/path")
            d.add_file_servers(123)

            assert "invalid file_server: 123" in str(e)

    def test_chaining(self):
        a = Directory(Directory.LOCAL_SCRATCH, "/path")
        b = a.add_file_servers(FileServer("url", Operation.PUT)).add_file_servers(
            FileServer("url", Operation.GET)
        )

        assert id(a) == id(b)

    def test_tojson(self):
        directory = Directory(Directory.LOCAL_SCRATCH, "/path", True).add_file_servers(
            FileServer("url", Operation.PUT)
        )

        result = json.loads(json.dumps(directory, cls=_CustomEncoder))

        expected = {
            "type": "localScratch",
            "path": "/path",
            "sharedFileSystem": True,
            "fileServers": [{"url": "url", "operation": "put"}],
        }

        assert result == expected


class TestGrid:
    def test_valid_grid(self):
        assert Grid(
            Grid.GT5,
            "smarty.isi.edu/jobmanager-pbs",
            Scheduler.PBS,
            SupportedJobs.AUXILLARY,
        )

    @pytest.mark.parametrize(
        "grid_type, contact, scheduler_type, job_type, invalid_var",
        [
            (
                "badgridtype",
                "contact",
                Scheduler.SLURM,
                SupportedJobs.AUXILLARY,
                "grid_type",
            ),
            (
                Grid.BATCH,
                "contact",
                "badschedulertype",
                SupportedJobs.AUXILLARY,
                "scheduler_type",
            ),
            (Grid.BATCH, "contact", Scheduler.SLURM, "badjobtype", "job_type"),
        ],
    )
    def test_invalid_grid(
        self, grid_type, contact, scheduler_type, job_type, invalid_var
    ):
        with pytest.raises(TypeError) as e:
            Grid(
                grid_type, contact, scheduler_type, job_type,
            )

        assert "invalid {invalid_var}: {value}".format(
            invalid_var=invalid_var, value=locals()[invalid_var]
        ) in str(e)

    def test_tojson(self, convert_yaml_schemas_to_json, load_schema):
        grid = Grid(
            Grid.GT5,
            "smarty.isi.edu/jobmanager-pbs",
            Scheduler.PBS,
            SupportedJobs.AUXILLARY,
        )

        result = json.loads(json.dumps(grid, cls=_CustomEncoder))

        expected = {
            "type": "gt5",
            "contact": "smarty.isi.edu/jobmanager-pbs",
            "scheduler": "pbs",
            "jobtype": "auxillary",
        }

        grid_schema = load_schema("sc-5.0.json")["$defs"]["grid"]
        validate(instance=result, schema=grid_schema)

        assert result == expected


class TestSite:
    @pytest.mark.parametrize(
        "arch", [(Arch.AARCH64), (Arch.X86_64),],
    )
    def test_valid_site(self, arch):
        assert Site(
            "site",
            arch=arch,
            os_type=OS.LINUX,
            os_release="release",
            os_version="1.1.1",
        )

    @pytest.mark.parametrize(
        "name, arch, os_type, invalid_var",
        [
            ("site", "badarch", OS.LINUX, "arch"),
            ("site", Arch.X86_64, "badostype", "os_type"),
            ("site", Arch.AARCH64, "randomos", "os_type"),
        ],
    )
    def test_invalid_site(self, name, arch, os_type, invalid_var):
        with pytest.raises(TypeError) as e:
            Site(name, arch=arch, os_type=os_type)

        assert "invalid {invalid_var}: {value}".format(
            invalid_var=invalid_var, value=locals()[invalid_var]
        ) in str(e)

    def test_add_valid_directory(self):
        site = Site("s")
        site.add_directories(Directory(Directory.LOCAL_SCRATCH, "/path"))
        site.add_directories(Directory(Directory.LOCAL_STORAGE, "/path"))

        assert len(site.directories) == 2

    def test_add_invalid_directory(self):
        with pytest.raises(TypeError) as e:
            site = Site("s")
            site.add_directories("baddirectory")

        assert "invalid directory: baddirectory" in str(e)

    def test_add_valid_grid(self):
        site = Site("s")
        site.add_grids(
            Grid(
                Grid.GT5,
                "smarty.isi.edu/jobmanager-pbs",
                Scheduler.PBS,
                job_type=SupportedJobs.AUXILLARY,
            )
        )
        site.add_grids(
            Grid(
                Grid.GT5,
                "smarty.isi.edu/jobmanager-pbs",
                Scheduler.PBS,
                job_type=SupportedJobs.COMPUTE,
            )
        )

        assert len(site.grids) == 2

    def test_add_invalid_grid(self):
        with pytest.raises(TypeError) as e:
            site = Site("s")
            site.add_grids("badgrid")

        assert "invalid grid: badgrid" in str(e)

    def test_chaining(self):
        site = Site("s")
        a = site.add_directories(Directory(Directory.LOCAL_SCRATCH, "/path"))
        b = site.add_grids(
            Grid(
                Grid.GT5,
                "smarty.isi.edu/jobmanager-pbs",
                Scheduler.PBS,
                job_type=SupportedJobs.AUXILLARY,
            )
        )

        assert id(a) == id(b)

    def test_tojson_with_profiles(self):
        site = Site(
            "s",
            arch=Arch.AARCH64,
            os_type=OS.LINUX,
            os_release="release",
            os_version="1.2.3",
        )
        site.add_directories(
            Directory(Directory.LOCAL_SCRATCH, "/path").add_file_servers(
                FileServer("url", Operation.GET)
            )
        )
        site.add_grids(
            Grid(
                Grid.GT5,
                "smarty.isi.edu/jobmanager-pbs",
                Scheduler.PBS,
                job_type=SupportedJobs.AUXILLARY,
            )
        )
        site.add_env(JAVA_HOME="/usr/bin/java")

        result = json.loads(json.dumps(site, cls=_CustomEncoder))

        expected = {
            "name": "s",
            "arch": "aarch64",
            "os.type": "linux",
            "os.release": "release",
            "os.version": "1.2.3",
            "directories": [
                {
                    "type": "localScratch",
                    "path": "/path",
                    "sharedFileSystem": False,
                    "fileServers": [{"url": "url", "operation": "get"}],
                }
            ],
            "grids": [
                {
                    "type": "gt5",
                    "contact": "smarty.isi.edu/jobmanager-pbs",
                    "scheduler": "pbs",
                    "jobtype": "auxillary",
                }
            ],
            "profiles": {"env": {"JAVA_HOME": "/usr/bin/java"}},
        }

        assert result == expected


@pytest.fixture
def expected_json():
    expected = {
        "pegasus": PEGASUS_VERSION,
        "sites": [
            {
                "name": "local",
                "arch": "x86_64",
                "os.type": "linux",
                "directories": [
                    {
                        "type": "sharedScratch",
                        "path": "/tmp/workflows/scratch",
                        "sharedFileSystem": False,
                        "fileServers": [
                            {
                                "url": "file:///tmp/workflows/scratch",
                                "operation": "all",
                            }
                        ],
                    },
                    {
                        "type": "localStorage",
                        "path": "/tmp/workflows/outputs",
                        "sharedFileSystem": False,
                        "fileServers": [
                            {
                                "url": "file:///tmp/workflows/outputs",
                                "operation": "all",
                            }
                        ],
                    },
                ],
            },
            {
                "name": "condor_pool",
                "arch": "x86_64",
                "os.type": "linux",
                "directories": [
                    {
                        "type": "sharedScratch",
                        "path": "/lustre",
                        "sharedFileSystem": False,
                        "fileServers": [
                            {
                                "url": "gsiftp://smarty.isi.edu/lustre",
                                "operation": "all",
                            }
                        ],
                    }
                ],
                "grids": [
                    {
                        "type": "gt5",
                        "contact": "smarty.isi.edu/jobmanager-pbs",
                        "scheduler": "pbs",
                        "jobtype": "auxillary",
                    },
                    {
                        "type": "gt5",
                        "contact": "smarty.isi.edu/jobmanager-pbs",
                        "scheduler": "pbs",
                        "jobtype": "compute",
                    },
                ],
                "profiles": {"env": {"JAVA_HOME": "/usr/bin/java"}},
            },
            {
                "name": "staging_site",
                "arch": "x86_64",
                "os.type": "linux",
                "directories": [
                    {
                        "type": "sharedScratch",
                        "path": "/data",
                        "sharedFileSystem": False,
                        "fileServers": [
                            {"url": "scp://obelix.isi.edu/data", "operation": "put",},
                            {"url": "http://obelix.isi.edu/data", "operation": "get",},
                        ],
                    }
                ],
            },
        ],
    }

    expected["sites"].sort(key=lambda s: s["name"])
    for i in range(len(expected["sites"])):
        expected["sites"][i]["directories"].sort(key=lambda d: d["path"])

        for j in range(len(expected["sites"][i]["directories"])):
            expected["sites"][i]["directories"][j]["fileServers"].sort(
                key=lambda fs: fs["url"]
            )

        if "grids" in expected["sites"][i]:
            expected["sites"][i]["grids"].sort(key=lambda g: g["jobtype"])

    return expected


class TestSiteCatalog:
    def test_add_valid_site(self):
        sc = SiteCatalog()
        assert sc.add_sites(Site("local"))

    def test_add_invalid_site(self):
        with pytest.raises(TypeError) as e:
            sc = SiteCatalog()
            sc.add_sites("badsite")

        assert "invalid site: badsite" in str(e)

    def test_add_duplicate_site(self):
        sc = SiteCatalog()
        sc.add_sites(Site("local"))
        with pytest.raises(DuplicateError):
            sc.add_sites(Site("local"))

    def test_chaining(self):
        sc = SiteCatalog()
        a = sc.add_sites(Site("local"))
        b = sc.add_sites(Site("condor_pool"))

        assert id(a) == id(b)

    def test_tojson(self, convert_yaml_schemas_to_json, load_schema, expected_json):
        sc = SiteCatalog().add_sites(
            Site("local", arch=Arch.X86_64, os_type=OS.LINUX).add_directories(
                Directory(
                    Directory.SHARED_SCRATCH, "/tmp/workflows/scratch"
                ).add_file_servers(
                    FileServer("file:///tmp/workflows/scratch", Operation.ALL)
                ),
                Directory(
                    Directory.LOCAL_STORAGE, "/tmp/workflows/outputs"
                ).add_file_servers(
                    FileServer("file:///tmp/workflows/outputs", Operation.ALL)
                ),
            ),
            Site("condor_pool", arch=Arch.X86_64, os_type=OS.LINUX)
            .add_directories(
                Directory(Directory.SHARED_SCRATCH, "/lustre").add_file_servers(
                    FileServer("gsiftp://smarty.isi.edu/lustre", Operation.ALL)
                )
            )
            .add_grids(
                Grid(
                    Grid.GT5,
                    "smarty.isi.edu/jobmanager-pbs",
                    Scheduler.PBS,
                    job_type=SupportedJobs.AUXILLARY,
                ),
                Grid(
                    Grid.GT5,
                    "smarty.isi.edu/jobmanager-pbs",
                    Scheduler.PBS,
                    job_type=SupportedJobs.COMPUTE,
                ),
            )
            .add_env(JAVA_HOME="/usr/bin/java"),
            Site("staging_site", arch=Arch.X86_64, os_type=OS.LINUX).add_directories(
                Directory(Directory.SHARED_SCRATCH, "/data")
                .add_file_servers(
                    FileServer("scp://obelix.isi.edu/data", Operation.PUT)
                )
                .add_file_servers(
                    FileServer("http://obelix.isi.edu/data", Operation.GET)
                )
            ),
        )

        result = json.loads(json.dumps(sc, cls=_CustomEncoder))

        sc_schema = load_schema("sc-5.0.json")
        validate(instance=result, schema=sc_schema)

        result["sites"].sort(key=lambda s: s["name"])
        for i in range(len(result["sites"])):
            result["sites"][i]["directories"].sort(key=lambda d: d["path"])

            for j in range(len(result["sites"][i]["directories"])):
                result["sites"][i]["directories"][j]["fileServers"].sort(
                    key=lambda fs: fs["url"]
                )

            if "grids" in result["sites"][i]:
                result["sites"][i]["grids"].sort(key=lambda g: g["jobtype"])

        assert result == expected_json

    @pytest.mark.parametrize(
        "_format, loader", [("json", json.load), ("yml", yaml.safe_load)]
    )
    def test_write(self, expected_json, _format, loader):
        sc = (
            SiteCatalog()
            .add_sites(
                Site("local", arch=Arch.X86_64, os_type=OS.LINUX).add_directories(
                    Directory(
                        Directory.SHARED_SCRATCH, "/tmp/workflows/scratch"
                    ).add_file_servers(
                        FileServer("file:///tmp/workflows/scratch", Operation.ALL)
                    ),
                    Directory(
                        Directory.LOCAL_STORAGE, "/tmp/workflows/outputs"
                    ).add_file_servers(
                        FileServer("file:///tmp/workflows/outputs", Operation.ALL)
                    ),
                )
            )
            .add_sites(
                Site("condor_pool", arch=Arch.X86_64, os_type=OS.LINUX)
                .add_directories(
                    Directory(Directory.SHARED_SCRATCH, "/lustre").add_file_servers(
                        FileServer("gsiftp://smarty.isi.edu/lustre", Operation.ALL)
                    )
                )
                .add_grids(
                    Grid(
                        Grid.GT5,
                        "smarty.isi.edu/jobmanager-pbs",
                        Scheduler.PBS,
                        job_type=SupportedJobs.AUXILLARY,
                    ),
                    Grid(
                        Grid.GT5,
                        "smarty.isi.edu/jobmanager-pbs",
                        Scheduler.PBS,
                        job_type=SupportedJobs.COMPUTE,
                    ),
                )
                .add_env(JAVA_HOME="/usr/bin/java"),
                Site(
                    "staging_site", arch=Arch.X86_64, os_type=OS.LINUX
                ).add_directories(
                    Directory(Directory.SHARED_SCRATCH, "/data").add_file_servers(
                        FileServer("scp://obelix.isi.edu/data", Operation.PUT),
                        FileServer("http://obelix.isi.edu/data", Operation.GET),
                    )
                ),
            )
        )

        with NamedTemporaryFile(mode="r+") as f:
            sc.write(f, _format=_format)
            f.seek(0)
            result = loader(f)

        result["sites"].sort(key=lambda s: s["name"])
        for i in range(len(result["sites"])):
            result["sites"][i]["directories"].sort(key=lambda d: d["path"])

            for j in range(len(result["sites"][i]["directories"])):
                result["sites"][i]["directories"][j]["fileServers"].sort(
                    key=lambda fs: fs["url"]
                )

            if "grids" in result["sites"][i]:
                result["sites"][i]["grids"].sort(key=lambda g: g["jobtype"])

        assert "createdOn" in result["x-pegasus"]
        assert result["x-pegasus"]["createdBy"] == getpass.getuser()
        assert result["x-pegasus"]["apiLang"] == "python"
        del result["x-pegasus"]
        assert result == expected_json

    def test_write_default(self):
        expected_file = Path("sites.yml")
        SiteCatalog().write()

        try:
            expected_file.unlink()
        except FileNotFoundError:
            pytest.fail(f"could not find {expected_file}")

    def test_site_catalog_key_ordering_on_yml_write(self):
        SiteCatalog().write()

        EXPECTED_FILE = Path("sites.yml")

        with EXPECTED_FILE.open() as f:
            # reading in as str so ordering of keys is not disrupted
            # when loaded into a dict
            result = f.read()

        EXPECTED_FILE.unlink()

        """
        Check that sc keys have been ordered as follows:
        - pegasus
        - sites
        """
        p = re.compile(r"x-pegasus:[\w\W]+pegasus: 5.0.4[\w\W]+sites:[\w\W]")
        assert p.match(result) is not None



from collections import OrderedDict, defaultdict
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

from .errors import DuplicateError
from .mixins import HookMixin, MetadataMixin, ProfileMixin
from .site_catalog import OS, Arch
from .writable import Writable, _filter_out_nones

from Pegasus.api._utils import _chained, _get_class_enum_member_str, _get_enum_str

PEGASUS_VERSION = "5.0.4"

__all__ = [
    "Container",
    "Transformation",
    "TransformationSite",
    "TransformationCatalog",
]


class _ContainerType(Enum):
    """Container types recognized by Pegasus"""

    DOCKER = "docker"
    SINGULARITY = "singularity"
    SHIFTER = "shifter"


class Container(ProfileMixin):
    """
    Describes a container that can be added to the :py:class:`~Pegasus.api.transformation_catalog.TransformationCatalog` .
    Note that the :code:`checksum` parameter refers to the checksum of the tar file of the image and not the specific
    version of an image (digest/content-addressable identifer in the case of Docker).

    .. code-block:: python

        # Example 1: Docker
        centos_pegasus = Container(
                            "centos-pegasus",
                            Container.DOCKER,
                            "docker:///ryan/centos-pegasus:latest",
                            arguments="--shm-size",
                            mounts=["/Volumes/Work/lfs1:/shared-data/:ro"]
                        )

        # Example 2: Singularity
        fb_nlp = Container(
                    "fb-nlp",
                    Container.SINGULARITY,
                    image="library://papajim/default/fb-nlp",
                    mounts=["/data:/mnt:ro"]
                )

    """

    DOCKER = _ContainerType.DOCKER
    SINGULARITY = _ContainerType.SINGULARITY
    SHIFTER = _ContainerType.SHIFTER

    _SUPPORTED_CHECKSUMS = {"sha256"}

    def __init__(
        self,
        name: str,
        container_type: _ContainerType,
        image: str,
        arguments: Optional[str] = None,
        mounts: Optional[List[str]] = None,
        image_site: Optional[str] = None,
        checksum: Optional[Dict[str, str]] = None,
        metadata: Optional[Dict[str, Union[int, str, float]]] = None,
        bypass_staging: bool = False,
    ):
        """
        :param name: name of this container
        :type name: str
        :param container_type: a container type defined in :py:class:`~Pegasus.api.transformation_catalog.Container`
        :type container_type: _ContainerType
        :param image: image, such as :code:`docker:///rynge/montage:latest`
        :type image: str
        :param arguments: additional cli arguments to be added to the :code:`docker container run` or :code:`singularity exec` commands when starting this container
        :type arguments: Optional[str]
        :param mounts: list of mount strings such as :code:`['/Volumes/Work/lfs1:/shared-data/:ro', ...]`
        :type mounts: Optional[List[str]]
        :param image_site: optional site attribute to tell pegasus which site tar file exists, defaults to None
        :type image_site: Optional[str]
        :param checksum: Dict containing checksums for the tar file of this image. Currently only sha256 is supported. This should be entered as :code:`{"sha256": <value>}`, defaults to None
        :type checksum: Optional[Dict[str, str]]
        :param metadata: Dict containing metadata key, value pairs associated with this container, defaults to None
        :type metadata: Optional[Dict[str, Union[int, str, float]]]
        :param bypass_staging: whether or not to bypass the stage in job for this container, defaults to False
        :type bypass_staging: bool, optional
        :raises ValueError: container_type must be one of :py:class:`~Pegasus.api.transformation_catalog._ContainerType` (:code:`Container.DOCKER` | :code:`Container.SINGULARITY` | :code:`Container.SHIFTER`)
        """
        self.name = name

        if not isinstance(container_type, _ContainerType):
            raise TypeError(
                "invalid container_type: {container_type}; container_type must be one of {enum_str}".format(
                    container_type=container_type,
                    enum_str=_get_class_enum_member_str(Container, _ContainerType),
                )
            )

        self.container_type = container_type.value
        self.image = image
        self.mounts = mounts
        self.image_site = image_site

        # ensure supported checksum type given
        if checksum and len(checksum) > 0:
            for checksum_type in checksum:
                if checksum_type.lower() not in Container._SUPPORTED_CHECKSUMS:
                    raise ValueError(
                        "invalid checksum: {}, supported checksum types are: {}".format(
                            checksum_type, Container._SUPPORTED_CHECKSUMS
                        )
                    )
        self.checksum = checksum

        # TODO: remove once this is supported
        if metadata:
            raise NotImplementedError(
                "Metadata support for Containers is not yet supported"
            )
        self.metadata = metadata

        self.profiles = defaultdict(OrderedDict)

        # add additional arguments if given (this is not part of the schema
        # and must be added to profiles)
        if arguments:
            self.add_pegasus_profile(container_arguments=arguments)

        self.bypass = None
        if bypass_staging:
            self.bypass = bypass_staging

    def __json__(self):
        return _filter_out_nones(
            OrderedDict(
                [
                    ("name", self.name),
                    ("type", self.container_type),
                    ("image", self.image),
                    ("mounts", self.mounts),
                    ("bypass", self.bypass),
                    ("image.site", self.image_site),
                    ("checksum", self.checksum),
                    ("metadata", self.metadata),
                    (
                        "profiles",
                        OrderedDict(sorted(self.profiles.items(), key=lambda _: _[0]))
                        if len(self.profiles) > 0
                        else None,
                    ),
                ]
            )
        )


class TransformationSite(ProfileMixin, MetadataMixin):
    """Site specific information about a :py:class:`~Pegasus.api.transformation_catalog.Transformation`.
    A :py:class:`~Pegasus.api.transformation_catalog.Transformation` must contain at least one
    transformation site.
    """

    def __init__(
        self,
        name: str,
        pfn: Union[str, Path],
        is_stageable: bool = False,
        bypass_staging: bool = False,
        arch: Optional[Arch] = None,
        os_type: Optional[OS] = None,
        os_release: Optional[str] = None,
        os_version: Optional[str] = None,
        container: Optional[Union[Container, str]] = None,
    ):
        """
        :param name: name of the site at which this :py:class:`~Pegasus.api.transformation_catalog.Transformation` resides
        :type name: str
        :param pfn: physical file name, an absolute path given as a str or Path
        :type pfn: Union[str, Path]
        :param is_stageable: whether or not this transformation is stageable or installed, defaults to False
        :type type: bool, optional
        :param bypass_staging: whether or not to bypass the stage in job of this executable (Note that this only works for transformations where :code:`is_stageable=True`), defaults to False
        :type bypass_staging: bool, optional
        :param arch: architecture that this :py:class:`~Pegasus.api.transformation_catalog.Transformation` was compiled for (defined in :py:class:`~Pegasus.api.site_catalog.Arch`), defaults to None
        :type arch: Optional[Arch], optional
        :param os_type: name of os that this :py:class:`~Pegasus.api.transformation_catalog.Transformation` was compiled for (defined in :py:class:`~Pegasus.api.site_catalog.OS`), defaults to None
        :type os_type: Optional[OS], optional
        :param os_release: release of os that this :py:class:`~Pegasus.api.transformation_catalog.Transformation` was compiled for, defaults to None, defaults to None
        :type os_release: Optional[str], optional
        :param os_version: version of os that this :py:class:`~Pegasus.api.transformation_catalog.Transformation` was compiled for, defaults to None, defaults to None
        :type os_version: Optional[str], optional
        :param container: specify the name of the container or Container object to use, optional
        :type container: Optional[Union[Container, str]], optional
        :raises TypeError: arch must be one of :py:class:`~Pegasus.api.site_catalog.Arch`
        :raises TypeError: os_type must be one of :py:class:`~Pegasus.api.site_catalog.OS`
        :raises ValueError: pfn must be given as an absolute path when :code:`pathlib.Path` is used
        :raises ValueError: :code:`bypass_staging=True` can only be used when :code:`is_stageable=True`
        """

        self.name = name

        if isinstance(pfn, Path):
            if not pfn.is_absolute():
                raise ValueError(
                    "invalid pfn: {}, the given pfn must be an absolute path".format(
                        str(pfn)
                    )
                )

            pfn = str(pfn)

        self.pfn = pfn
        self.transformation_type = "stageable" if is_stageable else "installed"

        self.bypass = None
        if bypass_staging:
            if not is_stageable:
                raise ValueError(
                    "bypass_staging can only be used when is_stageable is set to True"
                )

            self.bypass = bypass_staging

        if arch is not None:
            if not isinstance(arch, Arch):
                raise TypeError(
                    "invalid arch: {arch}; arch must be one of {enum_str}".format(
                        arch=arch, enum_str=_get_enum_str(Arch)
                    )
                )
            else:
                self.arch = arch.value
        else:
            self.arch = None

        if os_type is not None:
            if not isinstance(os_type, OS):
                raise TypeError(
                    "invalid os_type: {os_type}; os_type must be one of {enum_str}".format(
                        os_type=os_type, enum_str=_get_enum_str(OS)
                    )
                )
            else:
                self.os_type = os_type.value
        else:
            self.os_type = None

        self.os_release = os_release
        self.os_version = os_version

        container_name = None
        if container:
            if isinstance(container, Container):
                container_name = container.name
            elif isinstance(container, str):
                container_name = container
            else:
                raise TypeError(
                    "invalid container: {container}; container must be of type Container or str (the container name)".format(
                        container=container
                    )
                )
        self.container = container_name

        self.profiles = defaultdict(OrderedDict)
        self.metadata = OrderedDict()

    def __json__(self):
        return _filter_out_nones(
            OrderedDict(
                [
                    ("name", self.name),
                    ("pfn", self.pfn),
                    ("type", self.transformation_type),
                    ("bypass", self.bypass),
                    ("arch", self.arch),
                    ("os.type", self.os_type),
                    ("os.release", self.os_release),
                    ("os.version", self.os_version),
                    ("container", self.container),
                    (
                        "profiles",
                        OrderedDict(sorted(self.profiles.items(), key=lambda _: _[0]))
                        if len(self.profiles) > 0
                        else None,
                    ),
                    (
                        "metadata",
                        dict(self.metadata) if len(self.metadata) > 0 else None,
                    ),
                ]
            )
        )


class Transformation(ProfileMixin, HookMixin, MetadataMixin):
    """A transformation, which can be a standalone executable, or one that
    requires other executables.
    """

    _SUPPORTED_CHECKSUMS = {"sha256"}

    def __init__(
        self,
        name: str,
        namespace: Optional[str] = None,
        version: Optional[str] = None,
        site: Optional[str] = None,
        pfn: Optional[Union[str, Path]] = None,
        is_stageable: bool = False,
        bypass_staging: bool = False,
        arch: Optional[Arch] = None,
        os_type: Optional[OS] = None,
        os_release: Optional[str] = None,
        os_version: Optional[str] = None,
        container: Optional[Union[Container, str]] = None,
        checksum: Optional[Dict[str, str]] = None,
    ):
        """
        When a transformation resides on a single site, the
        syntax in Example 1 can be used where the args: site and pfn are
        provided to the constructor. If site and pfn are specified, then
        the args: is_stageable, bypass_staging, arch, os_type, os_release, os_version, and container, are
        applied to the site, else they are ignored. When a transformation resides
        multiple sites, the syntax in Example 2 can be used where multiple
        TransformationSite objects can be added. Note that when specifying a checksum
        such as :code:`{"sha256": <value>}` , this only applies stageable executables.

        .. code-block:: python

            # Example 1: transformation that resides on a single site
            preprocess = Transformation(
                    "preprocess",
                    namespace="pegasus",
                    version="4.0",
                    site="condorpool",
                    pfn="/usr/bin/pegasus-keg",
                    is_stageable=False,
                    bypass_staging=False,
                    arch=Arch.X86_64,
                    os_type=OS.LINUX,
                    container=centos_pegasus
                )

            # Example 2: transformation that resides on multiple sites
            preprocess = Transformation("preprocess", namespace="pegasus", version="4.0")\\
                            .add_sites(
                                TransformationSite(
                                    "condorpool",
                                    "/usr/bin/pegasus-keg",
                                    is_stageable=False,
                                    arch=Arch.X86_64,
                                    os_type=OS.LINUX,
                                    container="centos-pegasus"
                                ),
                                ...
                            )


        :param name: the logical name of the transformation
        :type name: str
        :param namespace: the namespace that this transformation belongs to, defaults to None
        :type namespace: Optional[str]
        :param version: the version of this transformation (e.g. :code:`"1.1"`), defaults to None
        :type version: Optional[str]
        :param site: a :py:class:`~Pegasus.api.site_catalog.Site` specified in the :py:class:`~Pegasus.api.site_catalog.SiteCatalog` on which this transformation resides, defaults to None
        :type site: Optional[str]
        :param pfn: the physical filename of this transformation (e.g. :code:`"/usr/bin/tar"`), defaults to None
        :type pfn: Optional[Union[str, Path]]
        :param is_stageable: whether or not this transformation is stageable or installed, defaults to False
        :type type: bool, optional
        :param bypass_staging: whether or not to bypass the stage in job of this executable (Note that this only works for transformations where :code:`is_stageable=True`), defaults to False
        :type bypass_staging: bool, optional
        :param arch: architecture that this transformation was compiled for (defined in :py:class:`~Pegasus.api.site_catalog.Arch` , e.g :code:`Arch.X86_64`), defaults to None
        :type arch: Optional[Arch]
        :param os_type: name of os that this transformation was compiled for (defined in :py:class:`~Pegasus.api.site_catalog.OS` , e.g. :code:`OS.LINUX`), defaults to None
        :type os_type: Optional[OS]
        :param os_release: release of os that this transformation was compiled for, defaults to None
        :type os_release: Optional[str]
        :param os_version: version of os that this transformation was compiled for, defaults to None
        :type os_version: Optional[str]
        :param container: a :py:class:`~Pegasus.api.transformation_catalog.Container` or name of the container to be used for this transformation, defaults to None
        :type container: Optional[Union[Container, str]]
        :param checksum: Dict containing checksums for this file. Currently only sha256 is given. This should be entered as :code:`{"sha256": <value>}`, defaults to None
        :type checksum: Optional[Dict[str, str]]
        :raises TypeError: container must be of type :py:class:`~Pegasus.api.transformation_catalog.Container` or str
        :raises TypeError: arch must be one of :py:class:`~Pegasus.api.site_catalog.Arch`
        :raises TypeError: os_type must be one of :py:class:`~Pegasus.api.site_catalog.OS`
        :raises ValueError: fields: namespace, name, and field must not contain any :code:`:` (colons)
        :raises ValueError: pfn must be given as an absolute path when :code:`pathlib.Path` is used
        :raises ValueError: :code:`bypass_staging=True` can only be used when :code:`is_stageable=True`
        """

        for field, value in {
            "name": name,
            "namespace": namespace,
            "version": version,
        }.items():
            if ":" in str(value):
                raise ValueError(
                    "invalid {field}: {value}; {field} must not contain ':' (colon) characters".format(
                        field=field, value=value
                    )
                )

        self.name = name
        self.namespace = namespace
        self.version = version

        self.sites = OrderedDict()
        self.requires = set()

        self.hooks = defaultdict(list)
        self.profiles = defaultdict(OrderedDict)
        self.metadata = OrderedDict()

        # add site if site if given
        if site is not None and pfn is not None:
            self.add_sites(
                TransformationSite(
                    site,
                    pfn,
                    is_stageable,
                    bypass_staging=bypass_staging,
                    arch=arch,
                    os_type=os_type,
                    os_release=os_release,
                    os_version=os_version,
                    container=container,
                )
            )

        # ensure supported checksum type given
        if checksum and len(checksum) > 0:
            for checksum_type in checksum:
                if checksum_type.lower() not in Transformation._SUPPORTED_CHECKSUMS:
                    raise ValueError(
                        "invalid checksum: {}, supported checksum types are: {}".format(
                            checksum_type, Transformation._SUPPORTED_CHECKSUMS
                        )
                    )
        self.checksum = checksum

    def _get_key(self):
        return f"{self.namespace}::{self.name}::{self.version}"

    @_chained
    def add_sites(self, *transformation_sites: TransformationSite):
        """
        add_sites(self, *transformation_sites: TransformationSite)
        Add one or more :py:class:`~Pegasus.api.transformation_catalog.TransformationSite` s to this
        transformation.

        :param transformation_sites: the :py:class:`~Pegasus.api.transformation_catalog.TransformationSite` (s) to be added
        :raises TypeError: argument(s) must be of type :py:class:`~Pegasus.api.transformation_catalog.TransformationSite`
        :raises DuplicateError: a :py:class:`~Pegasus.api.transformation_catalog.TransformationSite` with the same name as the one you are attempting to add already exists
        :return: self
        """
        for ts in transformation_sites:
            if not isinstance(ts, TransformationSite):
                raise TypeError(
                    "invalid transformation_site: {transformation_site}; transformation_site must be of type TransformationSite".format(
                        transformation_site=ts
                    )
                )

            if ts.name in self.sites:
                raise DuplicateError(
                    "transformation site: {name} has already been added to {transformation}".format(
                        name=ts.name, transformation=self
                    )
                )

            self.sites[ts.name] = ts

    @_chained
    def add_requirement(
        self,
        required_transformation: Union[str, "Transformation"],
        namespace: str = None,
        version: str = None,
    ):
        """
        add_requirement(self, required_transformation: Union[str, Transformation], namespace: str = None, version: str = None)
        Add a requirement to this transformation. Specify the other
        transformation, identified by name, namespace, and version, that this
        transformation depends upon. If a :py:class:`~Pegasus.api.transformation_catalog.Transformation`
        is passed in for :code:`required_transformation`, then namespace and version
        are ignored.

        :param required_transformation: :py:class:`~Pegasus.api.transformation_catalog.Transformation` that this transformation requires
        :type required_transformation: str or Transformation
        :raises DuplicateError: this requirement already exists
        :raises ValueError: :code:`required_transformation` must be of type :py:class:`~Pegasus.api.transformation_catalog.Transformation` or str
        :raises ValueError: namespace, required transformation name, and version cannot contain any :code:`:` (colon) characters
        :raises TypeError: required_transformation must be one of type str or :py:class:`~Pegasus.api.transformation_catalog.Transformation`
        :return: self
        """
        key = ""
        if isinstance(required_transformation, Transformation):
            if required_transformation.namespace:
                key += required_transformation.namespace + "::"

            key += required_transformation.name

            if required_transformation.version:
                key += ":" + required_transformation.version

        elif isinstance(required_transformation, str):
            for field, value in {
                "namespace": namespace,
                "required_transformation": required_transformation,
                "version": version,
            }.items():
                if ":" in str(value):
                    raise ValueError(
                        "invalid {field}: {value}; {field} must not contain `:` characters".format(
                            field=field, value=value
                        )
                    )

            if namespace:
                key += namespace + "::"

            key += required_transformation

            if version:
                key += ":" + version

        else:
            raise TypeError(
                "invalid required_transformation: {required_transformation}; required_transformation must be of type Transformation or str".format(
                    required_transformation=required_transformation
                )
            )

        if key in self.requires:
            raise DuplicateError(
                "transformation: {key} already added as a required transformation to {tr}".format(
                    key=key, tr=self
                )
            )

        self.requires.add(key)

    def __json__(self):
        return _filter_out_nones(
            OrderedDict(
                [
                    ("namespace", self.namespace),
                    ("name", self.name),
                    ("version", self.version),
                    (
                        "requires",
                        [req for req in self.requires]
                        if len(self.requires) > 0
                        else None,
                    ),
                    ("sites", [site for _, site in self.sites.items()]),
                    (
                        "profiles",
                        OrderedDict(sorted(self.profiles.items(), key=lambda _: _[0]))
                        if len(self.profiles) > 0
                        else None,
                    ),
                    (
                        "hooks",
                        OrderedDict(
                            [
                                (hook_name, [hook for hook in values])
                                for hook_name, values in self.hooks.items()
                            ]
                        )
                        if len(self.hooks) > 0
                        else None,
                    ),
                    ("metadata", self.metadata if len(self.metadata) > 0 else None),
                    ("checksum", self.checksum),
                ]
            )
        )

    def __str__(self):
        return "<Transformation {}::{}:{}>".format(
            self.namespace, self.name, self.version
        )

    def __hash__(self):
        return hash(self._get_key())

    def __eq__(self, other):
        if isinstance(other, Transformation):
            return self._get_key() == other._get_key()
        raise ValueError(f"Transformation cannot be compared with {type(other)}")


class TransformationCatalog(Writable):
    """Maintains a list a :py:class:`~Pegasus.api.transformation_catalog.Transformation` s and
    :py:class:~`Pegasus.api.transformation_catalog.Container` s

    .. code-block:: python

        # Example
        centos_pegasus = Container(
                "centos-pegasus",
                Container.DOCKER,
                "docker:///ryan/centos-pegasus:latest",
                mounts=["/Volumes/Work/lfs1:/shared-data/:ro"]
            )

        preprocess = Transformation(
                        "preprocess",
                        site="condorpool",
                        pfn="/usr/bin/pegasus-keg",
                        is_stageable=False,
                        arch=Arch.X86_64,
                        os_type=OS.LINUX,
                        container=centos_pegasus
                    )

        findrange = Transformation(
                        "findrange",
                        site="condorpool",
                        pfn="/usr/bin/pegasus-keg",
                        is_stageable=False,
                        arch=Arch.X86_64,
                        os_type=OS.LINUX,
                        container=centos_pegasus
                    )

        analyze = Transformation(
                        "analyze",
                        site="condorpool",
                        pfn="/usr/bin/pegasus-keg",
                        is_stageable=False,
                        arch=Arch.X86_64,
                        os_type=OS.LINUX,
                        container=centos_pegasus
                    )

        tc = TransformationCatalog()\\
            .add_containers(centos_pegasus)\\
            .add_transformations(preprocess, findrange, analyze)\\
            .write()

    """

    _DEFAULT_FILENAME = "transformations.yml"

    def __init__(self):
        Writable.__init__(self)

        self.transformations = OrderedDict()
        self.containers = OrderedDict()

    @_chained
    def add_transformations(self, *transformations: Transformation):
        """
        add_transformations(self, *transformations: Transformation)
        Add one or more :py:class:`~Pegasus.api.transformation_catalog.Transformation` s to this catalog

        .. code-block:: python

            # Example
            tc.add_transformations(
                Transformation(
                        "analyze",
                        site="condorpool",
                        pfn="/usr/bin/pegasus-keg",
                        is_stageable=False,
                        arch=Arch.X86_64,
                        os_type=OS.LINUX,
                        container=centos_pegasus
                    )
            )


        :param transformations: the :py:class:`~Pegasus.api.transformation_catalog.Transformation` (s) to be added
        :raises TypeError: argument(s) must be of type :py:class:`~Pegasus.api.transformation_catalog.Transformation`
        :raises DuplicateError: the given :py:class:`~Pegasus.api.transformation_catalog.Transformation` already exists in this catalog
        :return: self
        """
        for tr in transformations:
            if not isinstance(tr, Transformation):
                raise TypeError(
                    "invalid transformation: {tr}, transformation(s) must be of type Transformation".format(
                        tr=tr
                    )
                )

            if tr._get_key() in self.transformations:
                raise DuplicateError(
                    "transformation: {key} has already been added to this TransformationCatalog".format(
                        key=tr._get_key()
                    )
                )

            self.transformations[tr._get_key()] = tr

    @_chained
    def add_containers(self, *containers: Container):
        """
        add_containers(self, *containers: Container)
        Add one or more :py:class:`~Pegasus.api.transformation_catalog.Container` s to this catalog

        .. code-block:: python

            # Example
            tc.add_containers(
                Container(
                    "centos-pegasus",
                    Container.DOCKER,
                    "docker:///ryan/centos-pegasus:latest",
                    mounts=["/Volumes/Work/lfs1:/shared-data/:ro"]
                )
            )

        :param containers: the :py:class:`~Pegasus.api.transformation_catalog.Container` to be added
        :raises TypeError: argument(s) must be of type :py:class:`~Pegasus.api.transformation_catalog.Container`
        :raises DuplicateError: a container with the same name already exists in this catalog
        :return: self

        """
        for c in containers:
            if not isinstance(c, Container):
                raise TypeError(
                    "invalid container: {container}; container must be of type Container".format(
                        container=c
                    )
                )

            if c.name in self.containers:
                raise DuplicateError(
                    "container: {} has already been added to this TransformationCatalog".format(
                        c.name
                    )
                )

            self.containers[c.name] = c

    def __json__(self):
        containers = None
        if len(self.containers) > 0:
            containers = [c for _, c in self.containers.items()]

        return _filter_out_nones(
            OrderedDict(
                [
                    ("pegasus", PEGASUS_VERSION),
                    ("transformations", [t for _, t in self.transformations.items()]),
                    ("containers", containers),
                ]
            )
        )
